from __future__ import unicode_literals
from builtins import object
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.tools import numerix

class _SparseMatrix(object):

    """
    .. attention:: This class is abstract. Always create one of its subclasses.
    """

    matrix = None
    numpyArray = property()
    _shape = property()

    def __init__(self):
        pass

    __array_priority__ = 100.0

    def __array_wrap(self, arr, context=None):
        if context is None:
            return arr
        else:
            return NotImplemented

    def copy(self):
        raise NotImplementedError

    def __getitem__(self, index):
        raise NotImplementedError

    @property
    def _range(self):
        raise NotImplementedError

    def __str__(self):
        s = ''
        cellWidth = 11
        Irange, Jrange = self._range

        for j in Jrange:
            for i in Irange:
                v = self[j, i]
                if v == 0:
                    s += "---".center(cellWidth)
                else:
                    exp = numerix.log(abs(v))
                    if abs(exp) <= 4:
                        if exp < 0:
                            s += ("%9.6f" % v).ljust(cellWidth)
                        else:
                            s += ("%9.*f" % (6, v)).ljust(cellWidth)
                    else:
                        s += ("%9.2e" % v).ljust(cellWidth)
            s += "\n"
        return s[:-1]

    def __repr__(self):
        return repr(self.matrix)

    def __setitem__(self, index, value):
        raise NotImplementedError

    def __add__(self, other):
        raise NotImplementedError

    __radd__ = __add__

    def __iadd__(self, other):
        raise NotImplementedError

    def __sub__(self, other):
        raise NotImplementedError

    # Ask about this rsub
    def __rsub__(self, other):
        return -(self.__sub__(other))


    def __isub__(self, other):
        raise NotImplementedError

    def __mul__(self, other):
        raise NotImplementedError

    def __rmul__(self, other):
        raise NotImplementedError

    def __neg__(self):
        return self * -1

    def __pos__(self):
        return self

##     def __eq__(self,other):
## 	return self.matrix.__eq__(other.matrix)

    def put(self, vector, id1, id2):
        raise NotImplementedError

    def putDiagonal(self, vector):
        raise NotImplementedError

    def take(self, id1, id2):
        raise NotImplementedError

    def takeDiagonal(self):
        raise NotImplementedError

    def addAt(self, vector, id1, id2):
        raise NotImplementedError

    def addAtDiagonal(self, vector):
        raise NotImplementedError

    def exportMmf(self, filename):
        raise NotImplementedError

    @property
    def CSR(self):
        raise NotImplementedError

    @property
    def LIL(self):
        raise NotImplementedError

    @property
    def T(self):
        raise NotImplementedError

    def _matrix2mesh(self, ids):
        """Convert matrix row indices to mesh cell indices
        """
        return ids

    def _mesh2matrix(self, ids):
        """Convert mesh cell indices to matrix row indices
        """
        return ids

##     def __array__(self):
##      shape = self._shape
##      indices = numerix.indices(shape)
##         numMatrix = self.take(indices[0].ravel(), indices[1].ravel())
##      return numerix.reshape(numMatrix, shape)

class _Mesh2Matrix(object):
    _bodies = None
    _ghosts = None

    def __init__(self, mesh, matrix,
                 numberOfEquations=1, numberOfVariables=1):
        """Creates a mapping between mesh cells and matrix rows and columns

        Parameters
        ----------
        mesh : ~fipy.meshes.mesh.Mesh
            The `Mesh` to be mapped to a `_SparseMatrix`
        matrix : ~fipy.matrices.sparseMatrix.SparseMatrix
            The `_SparseMatrix` to be mapped to a `Mesh`
        numberOfVariables : int
            The local columns of the matrix are determined by
            `numberOfVariables * len(mesh._localNonOverlappingCellIDs)`.
        numberOfEquations : int
            The local rows of the matrix are determined by
            `numberOfEquations * len(mesh._localNonOverlappingCellIDs)`.
        """
        self.mesh = mesh
        self.numberOfVariables = numberOfVariables
        self.numberOfEquations = numberOfEquations

        # Note: storing matrix directly results in a reference cycle
        # between _PETScBaseMeshMatrix and _Mesh2Matrix, specifically
        # between _PETScBaseMeshMatrix.matrix and _PETScBaseMeshMatrix._ao.
        #
        # See https://lists.mcs.anl.gov/pipermail/petsc-users/2020-October/042652.html
        # and https://github.com/usnistgov/fipy/pull/761
        import weakref
        self.matrix = weakref.ref(matrix)

    @staticmethod
    def _cellIDsToGlobalIDs(IDs, M, L):
        N = len(IDs)
        return (numerix.vstack([IDs] * M) + numerix.indices((M, N))[0] * L).flatten()

    def _cellIDsToGlobalRowIDs(self, IDs):
        return self._cellIDsToGlobalIDs(IDs, M=self.numberOfEquations,
                                        L=self.mesh.globalNumberOfCells)

    def _cellIDsToLocalRowIDs(self, IDs):
        return self._cellIDsToGlobalIDs(IDs, M=self.numberOfEquations,
                                        L=self.mesh.numberOfCells)

    @property
    def globalNonOverlappingRowIDs(self):
        return self._cellIDsToGlobalRowIDs(self.mesh._globalNonOverlappingCellIDs)

    @property
    def globalOverlappingRowIDs(self):
        return self._cellIDsToGlobalRowIDs(self.mesh._globalOverlappingCellIDs)

    @property
    def localNonOverlappingRowIDs(self):
        return self._cellIDsToLocalRowIDs(self.mesh._localNonOverlappingCellIDs)

    def _cellIDsToGlobalColIDs(self, IDs):
        return self._cellIDsToGlobalIDs(IDs, M=self.numberOfVariables,
                                        L=self.mesh.globalNumberOfCells)

    def _cellIDsToLocalColIDs(self, IDs):
        return self._cellIDsToGlobalIDs(IDs, M=self.numberOfVariables,
                                        L=self.mesh.numberOfCells)

    @property
    def globalNonOverlappingColIDs(self):
        return self._cellIDsToGlobalColIDs(self.mesh._globalNonOverlappingCellIDs)

    @property
    def globalOverlappingColIDs(self):
        return self._cellIDsToGlobalColIDs(self.mesh._globalOverlappingCellIDs)

    @property
    def localOverlappingColIDs(self):
        return self._cellIDsToLocalColIDs(self.mesh._localOverlappingCellIDs)

    @property
    def localNonOverlappingColIDs(self):
        return self._cellIDsToLocalColIDs(self.mesh._localNonOverlappingCellIDs)

    def _getStencil_(self, id1, id2,
                     globalOverlappihgIDs, globalNonOverlappihgIDs,
                     overlapping=False):
        id1 = globalOverlappihgIDs[id1]

        if overlapping:
            mask = numerix.ones(id1.shape, dtype=bool)
        else:
            mask = numerix.in1d(id1, globalNonOverlappihgIDs)

        id1 = self.matrix()._mesh2matrix(id1[mask])
        id2 = numerix.asarray(id2)[mask]

        return id1, id2, mask

    def _getStencil(self, id1, id2, overlapping=False):
        raise NotImplementedError

    def globalVectorAndIDs(self, vector, id1, id2, overlapping=False):
        """Transforms local overlapping values and coordinates to global

        Parameters
        ----------
        vector : array_like
            The overlapping values to insert.
        id1 : array_like
            The local overlapping row indices.
        id2 : array_like
            The local overlapping column indices.
        overlapping : bool
            Whether to return ghosted values or not (default False)

        Returns
        -------
        tuple of ndarray
            (vector, global row indices, global column indices)
        """
        id1, id2, mask = self._getStencil(id1, id2, overlapping)
        vector = vector[mask]
        return (vector, id1, id2)

    @property
    def bodies(self):
        if self._bodies is None:
            self._bodies = numerix.in1d(self.mesh._globalOverlappingCellIDs,
                                        self.mesh._globalNonOverlappingCellIDs)
        return self._bodies

    @property
    def ghosts(self):
        if self._ghosts is None:
            self._ghosts = self.mesh._globalOverlappingCellIDs[~self.bodies]
            self._ghosts = self._cellIDsToGlobalRowIDs(self._ghosts)
            self._ghosts = self.matrix()._mesh2matrix(self._ghosts)

        return self._ghosts

class _RowMesh2Matrix(_Mesh2Matrix):
    def _getStencil(self, id1, id2, overlapping=False):
        return self._getStencil_(id1, id2,
                                 self.globalOverlappingRowIDs,
                                 self.globalNonOverlappingRowIDs,
                                 overlapping)

class _ColMesh2Matrix(_Mesh2Matrix):
    def _getStencil(self, id1, id2, overlapping=False):
        id2, id1, mask = self._getStencil_(id2, id1,
                                           self.globalOverlappingColIDs,
                                           self.globalNonOverlappingColIDs,
                                           overlapping)

        return id1, id2, mask

class _RowColMesh2Matrix(_RowMesh2Matrix):
    def _getStencil(self, id1, id2, overlapping=False):
        id2 = self.globalOverlappingColIDs[id2]

        id1, id2, mask = super(_RowColMesh2Matrix, self)._getStencil(id1, id2, overlapping)

        id2 = self.matrix()._mesh2matrix(id2)

        return id1, id2, mask

    def _test(self):
        """Tests

        >>> from fipy import Grid1D
        >>> from fipy.tools.numerix import allequal

        >>> m2m = _RowColMesh2Matrix(mesh=Grid1D(nx=5),
        ...                          numberOfVariables=3,
        ...                          numberOfEquations=2)
        >>> GOC = m2m.globalOverlappingColIDs
        >>> GNOC = m2m.globalNonOverlappingColIDs
        >>> LNOC = m2m.localNonOverlappingColIDs
        >>> GOR = m2m.globalOverlappingRowIDs
        >>> GNOR = m2m.globalNonOverlappingRowIDs
        >>> LNOR = m2m.localNonOverlappingRowIDs

        5 cells, 3 variables, 1 processor

        ```
        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   cell IDs
        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   column IDs

        _globalOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   proc 0

        globalOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   proc 0

        _globalNonOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   proc 0

        globalNonOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   proc 0

        _localOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   proc 0

        localOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   proc 0

        _localNonOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   proc 0

        localNonOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   proc 0
        ```

        >>> print(allequal(GOC, [0, 1, 2, 3, 4, 5, 6, 7,
        ...                      8, 9, 10, 11, 12, 13, 14]))  # doctest: +SERIAL
        True
        >>> print(allequal(GNOC, [0, 1, 2, 3, 4, 5, 6, 7,
        ...                       8, 9, 10, 11, 12, 13, 14])) # doctest: +SERIAL
        True
        >>> print(allequal(LNOC, [0, 1, 2, 3, 4, 5, 6, 7,
        ...                       8, 9, 10, 11, 12, 13, 14])) # doctest: +SERIAL
        True


        5 cells, 2 equations, 1 processor

        ```
        0  1  2  3  4  0  1  2  3  4   cell IDs
        0  1  2  3  4  5  6  7  8  9   row IDs

        _globalOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4   proc 0

        globalOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9   proc 0

        _globalNonOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4   proc 0

        globalNonOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9   proc 0

        _localOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4   proc 0

        _localOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9   proc 0

        _localNonOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4   proc 0

        localNonOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9   proc 0
        ```

        >>> print(allequal(GOR, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])) # doctest: +SERIAL
        True
        >>> print(allequal(GNOR, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])) # doctest: +SERIAL
        True
        >>> print(allequal(LNOR, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])) # doctest: +SERIAL
        True


        5 cells, 3 variables, 2 processors

        ```
        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   cell IDs
        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   column IDs

        _globalOverlappingCellIDs

        0  1  2  3     0  1  2  3     0  1  2  3      proc 0
        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   proc  1

        globalOverlappingColIDs

        0  1  2  3     5  6  7  8    10 11 12 13      proc 0
        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   proc  1

        _globalNonOverlappingCellIDs

        0  1           0  1           0  1            proc 0
              2  3  4        2  3  4        2  3  4   proc  1

        globalNonOverlappingColIDs

        0  1           5  6          10 11            proc 0
              2  3  4        7  8  9       12 13 14   proc  1

        _localOverlappingCellIDs

        0  1  2  3     0  1  2  3     0  1  2  3      proc 0
        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   proc  1

        localOverlappingColIDs

        0  1  2  3     4  5  6  7     8  9 10 11      proc 0
        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   proc  1

        _localNonOverlappingCellIDs

        0  1           0  1           0  1            proc 0
              2  3  4        2  3  4        2  3  4   proc  1

        localNonOverlappingColIDs

        0  1           4  5           8  9            proc 0
              2  3  4        7  8  9       12 13 14   proc  1
        ```


        >>> print(allequal(GOC, [0, 1, 2, 3, 5, 6, 7,
        ...                      8, 10, 11, 12, 13])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(GOC, [0, 1, 2, 3, 4, 5, 6, 7,
        ...                      8, 9, 10, 11, 12, 13, 14])) # doctest: +PROCESSOR_1_OF_2
        True

        >>> print(allequal(GNOC, [0, 1, 5, 6, 10, 11])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(GNOC, [2, 3, 4, 7, 8, 9,
        ...                       12, 13, 14])) # doctest: +PROCESSOR_1_OF_2
        True

        >>> print(allequal(LNOC, [0, 1, 4, 5, 8, 9])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(LNOC, [2, 3, 4, 7, 8, 9,
        ...                       12, 13, 14])) # doctest: +PROCESSOR_1_OF_2
        True


        5 cells, 2 equations, 2 processors

        ```
        0  1  2  3  4  0  1  2  3  4   cell IDs
        0  1  2  3  4  5  6  7  8  9   row IDs

        _globalOverlappingCellIDs

        0  1  2  3     0  1  2  3      proc 0
        0  1  2  3  4  0  1  2  3  4   proc  1

        globalOverlappingRowIDs

        0  1  2  3     5  6  7  8      proc 0
        0  1  2  3  4  5  6  7  8  9   proc  1

        _globalNonOverlappingCellIDs

        0  1           0  1            proc 0
              2  3  4        2  3  4   proc  1

        globalNonOverlappingRowIDs

        0  1           5  6            proc 0
              2  3  4        7  8  9   proc  1

        _localOverlappingCellIDs

        0  1  2  3     0  1  2  3      proc 0
        0  1  2  3  4  0  1  2  3  4   proc  1

        _localOverlappingRowIDs

        0  1  2  3     4  5  6  7      proc 0
        0  1  2  3  4  5  6  7  8  9   proc  1

        _localNonOverlappingCellIDs

        0  1           0  1            proc 0
              2  3  4        2  3  4   proc  1

        localNonOverlappingRowIDs

        0  1           4  5            proc 0
              2  3  4        7  8  9   proc  1
        ```


        >>> print(allequal(GOR, [0, 1, 2, 3, 5, 6, 7, 8])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(GOR, [0, 1, 2, 3, 4, 5,
        ...                      6, 7, 8, 9])) # doctest: +PROCESSOR_1_OF_2
        True

        >>> print(allequal(GNOR, [0, 1, 5, 6])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(GNOR, [2, 3, 4, 7, 8, 9])) # doctest: +PROCESSOR_1_OF_2
        True

        >>> print(allequal(LNOR, [0, 1, 4, 5])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(LNOR, [2, 3, 4, 7, 8, 9])) # doctest: +PROCESSOR_1_OF_2
        True

        >>> m2m = _RowColMesh2Matrix(mesh=Grid1D(nx=5, communicator=serialComm),
        ...                          numberOfVariables=3,
        ...                          numberOfEquations=2)
        >>> GOC = m2m.globalOverlappingColIDs
        >>> GNOC = m2m.globalNonOverlappingColIDs
        >>> LNOC = m2m.localNonOverlappingColIDs
        >>> GOR = m2m.globalOverlappingRowIDs
        >>> GNOR = m2m.globalNonOverlappingRowIDs
        >>> LNOR = m2m.localNonOverlappingRowIDs

        5 cells, 3 variables, serial

        ```
        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   cell IDs
        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   column IDs

        _globalOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   proc 0

        globalOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   proc 0

        _globalNonOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   proc 0

        globalNonOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   proc 0

        _localOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   proc 0

        localOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   proc 0

        _localNonOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4  0  1  2  3  4   proc 0

        localNonOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14   proc 0
        ```

        >>> print(allequal(GOC, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]))
        True
        >>> print(allequal(GNOC, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]))
        True
        >>> print(allequal(LNOC, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]))
        True


        5 cells, 2 equations, serial

        ```
        0  1  2  3  4  0  1  2  3  4   cell IDs
        0  1  2  3  4  5  6  7  8  9   row IDs

        _globalOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4   proc 0

        globalOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9   proc 0

        _globalNonOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4   proc 0

        globalNonOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9   proc 0

        _localOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4   proc 0

        _localOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9   proc 0

        _localNonOverlappingCellIDs

        0  1  2  3  4  0  1  2  3  4   proc 0

        localNonOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9   proc 0
        ```

        >>> print(allequal(GOR, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]))
        True
        >>> print(allequal(GNOR, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]))
        True
        >>> print(allequal(LNOR, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]))
        True

        >>> m2m = _RowColMesh2Matrix(mesh=Grid1D(nx=7),
        ...                          numberOfVariables=3,
        ...                          numberOfEquations=2)
        >>> GOC = m2m.globalOverlappingColIDs
        >>> GNOC = m2m.globalNonOverlappingColIDs
        >>> LNOC = m2m.localNonOverlappingColIDs
        >>> GOR = m2m.globalOverlappingRowIDs
        >>> GNOR = m2m.globalNonOverlappingRowIDs
        >>> LNOR = m2m.localNonOverlappingRowIDs

        7 cells, 3 variables, 1 processor

        ```
        0  1  2  3  4  5  6  0  1  2  3  4  5  6  0  1  2  3  4  5  6   cell IDs
        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20   column IDs

        _globalOverlappingCellIDs

        0  1  2  3  4  5  6  0  1  2  3  4  5  6  0  1  2  3  4  5  6   proc 0

        globalOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20   proc 0

        _globalNonOverlappingCellIDs

        0  1  2  3  4  5  6  0  1  2  3  4  5  6  0  1  2  3  4  5  6   proc 0

        globalNonOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20   proc 0

        _localOverlappingCellIDs

        0  1  2  3  4  5  6  0  1  2  3  4  5  6  0  1  2  3  4  5  6   proc 0

        localOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20   proc 0

        _localNonOverlappingCellIDs

        0  1  2  3  4  5  6  0  1  2  3  4  5  6  0  1  2  3  4  5  6   proc 0

        localNonOverlappingColIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20   proc 0
        ```

        >>> print(allequal(GOC, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11,
        ...                      12, 13, 14, 15, 16, 17, 18, 19, 20])) # doctest: +SERIAL
        True
        >>> print(allequal(GNOC, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11,
        ...                       12, 13, 14, 15, 16, 17, 18, 19, 20])) # doctest: +SERIAL
        True
        >>> print(allequal(LNOC, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11,
        ...                       12, 13, 14, 15, 16, 17, 18, 19, 20])) # doctest: +SERIAL
        True


        7 cells, 2 equations, 1 processor

        ```
        0  1  2  3  4  5  6  0  1  2  3  4  5  6   cell IDs
        0  1  2  3  4  5  6  7  8  9 10 11 12 13   row IDs

        _globalOverlappingCellIDs

        0  1  2  3  4  5  6  0  1  2  3  4  5  6   proc 0

        globalOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13   proc 0

        _globalNonOverlappingCellIDs

        0  1  2  3  4  5  6  0  1  2  3  4  5  6   proc 0

        globalNonOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13   proc 0

        _localOverlappingCellIDs

        0  1  2  3  4  5  6  0  1  2  3  4  5  6   proc 0

        _localOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13   proc 0

        _localNonOverlappingCellIDs

        0  1  2  3  4  5  6  0  1  2  3  4  5  6   proc 0

        localNonOverlappingRowIDs

        0  1  2  3  4  5  6  7  8  9 10 11 12 13   proc 0
        ```

        >>> print(allequal(GOR, [0, 1, 2, 3, 4, 5, 6, 7,
        ...                      8, 9, 10, 11, 12, 13])) # doctest: +SERIAL
        True
        >>> print(allequal(GNOR, [0, 1, 2, 3, 4, 5, 6, 7,
        ...                       8, 9, 10, 11, 12, 13])) # doctest: +SERIAL
        True
        >>> print(allequal(LNOR, [0, 1, 2, 3, 4, 5, 6, 7,
        ...                       8, 9, 10, 11, 12, 13])) # doctest: +SERIAL
        True


        7 cells, 3 variables, 2 processors

        ```
        0  1  2  3  4  5  6  0  1  2  3  4  5  6  0  1  2  3  4  5  6   cell IDs
        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20   column IDs

        _globalOverlappingCellIDs

        0  1  2  3  4        0  1  2  3  4        0  1  2  3  4         proc 0
           1  2  3  4  5  6     1  2  3  4  5  6     1  2  3  4  5  6   proc  1

        globalOverlappingColIDs

        0  1  2  3  4        7  8  9 10 11       14 15 16 17 18         proc 0
           1  2  3  4  5  6     8  9 10 11 12 13    15 16 17 18 19 20   proc  1

        _globalNonOverlappingCellIDs

        0  1  2              0  1  2              0  1  2               proc 0
                 3  4  5  6           3  4  5  6           3  4  5  6   proc  1

        globalNonOverlappingColIDs

        0  1  2              7  8  9             14 15 16               proc 0
                 3  4  5  6          10 11 12 13          17 18 19 20   proc  1

        _localOverlappingCellIDs

        0  1  2  3  4        0  1  2  3  4        0  1  2  3  4         proc 0
           0  1  2  3  4  5     0  1  2  3  4  5     0  1  2  3  4  5   proc  1

        localOverlappingColIDs

        0  1  2  3  4        5  6  7  8  9       10 11 12 13 14         proc 0
           0  1  2  3  4  5     6  7  8  9 10 11    12 13 14 15 16 17   proc  1

        _localNonOverlappingCellIDs

        0  1  2              0  1  2              0  1  2               proc 0
                 2  3  4  5           2  3  4  5           2  3  4  5   proc  1

        localNonOverlappingColIDs

        0  1  2              5  6  7             10 11 12               proc 0
                 2  3  4  5           8  9 10 11          14 15 16 17   proc  1
        ```

        >>> print(allequal(GOC, [0, 1, 2, 3, 4, 7, 8, 9, 10,
        ...                      11, 14, 15, 16, 17, 18])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(GOC, [1, 2, 3, 4, 5, 6, 8, 9, 10, 11, 12,
        ...                      13, 15, 16, 17, 18, 19, 20])) # doctest: +PROCESSOR_1_OF_2
        True

        >>> print(allequal(GNOC, [0, 1, 2, 7, 8,
        ...                       9, 14, 15, 16])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(GNOC, [3, 4, 5, 6, 10, 11, 12,
        ...                       13, 17, 18, 19, 20])) # doctest: +PROCESSOR_1_OF_2
        True

        >>> print(allequal(LNOC, [0, 1, 2, 5, 6,
        ...                       7, 10, 11, 12])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(LNOC, [2, 3, 4, 5, 8, 9, 10,
        ...                       11, 14, 15, 16, 17])) # doctest: +PROCESSOR_1_OF_2
        True

        7 cells, 2 equations, 2 processors

        ```
        0  1  2  3  4  5  6  0  1  2  3  4  5  6   cell IDs
        0  1  2  3  4  5  6  7  8  9 10 11 12 13   row IDs

        _globalOverlappingCellIDs

        0  1  2  3  4        0  1  2  3  4         proc 0
           1  2  3  4  5  6     1  2  3  4  5  6   proc  1

        globalOverlappingRowIDs

        0  1  2  3  4        7  8  9 10 11         proc 0
           1  2  3  4  5  6     8  9 10 11 12 13   proc  1

        _globalNonOverlappingCellIDs

        0  1  2              0  1  2               proc 0
                 3  4  5  6           3  4  5  6   proc  1

        globalNonOverlappingRowIDs

        0  1  2              7  8  9               proc 0
                 3  4  5  6          10 11 12 13   proc  1

        _localOverlappingCellIDs

        0  1  2  3  4        0  1  2  3  4         proc 0
           0  1  2  3  4  5     0  1  2  3  4  5   proc  1

        _localOverlappingRowIDs

        0  1  2  3  4        5  6  7  8  9         proc 0
           0  1  2  3  4  5     6  7  8  9 10 11   proc  1

        _localNonOverlappingCellIDs

        0  1  2              0  1  2               proc 0
                 2  3  4  5           2  3  4  5   proc  1

        localNonOverlappingRowIDs

        0  1  2              5  6  7               proc 0
                 2  3  4  5           8  9 10 11   proc  1
        ```

        >>> print(allequal(GOR, [0, 1, 2, 3, 4, 7,
        ...                      8, 9, 10, 11])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(GOR, [1, 2, 3, 4, 5, 6, 8, 9,
        ...                      10, 11, 12, 13])) # doctest: +PROCESSOR_1_OF_2
        True

        >>> print(allequal(GNOR, [0, 1, 2, 7, 8, 9])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(GNOR, [3, 4, 5, 6, 10, 11, 12, 13])) # doctest: +PROCESSOR_1_OF_2
        True

        >>> print(allequal(LNOR, [0, 1, 2, 5, 6, 7])) # doctest: +PROCESSOR_0_OF_2
        True
        >>> print(allequal(LNOR, [2, 3, 4, 5, 8, 9, 10, 11])) # doctest: +PROCESSOR_1_OF_2
        True


        7 cells, 3 variables, 3 processors

        ```
        0  1  2  3  4  5  6  0  1  2  3  4  5  6  0  1  2  3  4  5  6   cell IDs
        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20   column IDs

        _globalOverlappingCellIDs

        0  1  2  3           0  1  2  3           0  1  2  3            proc 0
        0  1  2  3  4  5     0  1  2  3  4  5     0  1  2  3  4  5      proc  1
              2  3  4  5  6        2  3  4  5  6        2  3  4  5  6   proc   2

        globalOverlappingColIDs

        0  1  2  3           7  8  9 10          14 15 16 17            proc 0
        0  1  2  3  4  5     7  8  9 10 11 12    14 15 16 17 18 19      proc  1
              2  3  4  5  6        9 10 11 12 13       16 17 18 19 20   proc   2

        _globalNonOverlappingCellIDs

        0  1                 0  1                 0  1                  proc 0
              2  3                 2  3                 2  3            proc  1
                    4  5  6              4  5  6              4  5  6   proc   2

        globalNonOverlappingColIDs

        0  1                 7  8                14 15                  proc 0
              2  3                 9 10                16 17            proc  1
                    4  5  6             11 12 13             18 19 20   proc   2

        _localOverlappingCellIDs

        0  1  2  3           0  1  2  3           0  1  2  3            proc 0
        0  1  2  3  4  5     0  1  2  3  4  5     0  1  2  3  4  5      proc  1
              0  1  2  3  4        0  1  2  3  4        0  1  2  3  4   proc   2

        localOverlappingColIDs

        0  1  2  3           4  5  6  7           8  9 10 11            proc 0
        0  1  2  3  4  5     6  7  8  9 10 11    12 13 14 15 16 17      proc  1
              0  1  2  3  4        5  6  7  8  9       10 11 12 13 14   proc   2

        _localNonOverlappingCellIDs

        0  1                 0  1                 0  1                  proc 0
              2  3                 2  3                 2  3            proc  1
                    2  3  4              2  3  4              2  3  4   proc   2

        localNonOverlappingColIDs

        0  1                 4  5                 8  9                  proc 0
              2  3                 8  9                14 15            proc  1
                    2  3  4              7  8  9             12 13 14   proc   2
        ```

        >>> print(allequal(GOC, [0, 1, 2, 3, 7, 8, 9,
        ...                      10, 14, 15, 16, 17])) # doctest: +PROCESSOR_0_OF_3
        True
        >>> print(allequal(GOC, [0, 1, 2, 3, 4, 5, 7, 8, 9, 10, 11,
        ...                      12, 14, 15, 16, 17, 18, 19])) # doctest: +PROCESSOR_1_OF_3
        True
        >>> print(allequal(GOC, [2, 3, 4, 5, 6, 9, 10, 11, 12,
        ...                      13, 16, 17, 18, 19, 20])) # doctest: +PROCESSOR_2_OF_3
        True

        >>> print(allequal(GNOC, [0, 1, 7, 8, 14, 15])) # doctest: +PROCESSOR_0_OF_3
        True
        >>> print(allequal(GNOC, [2, 3, 9, 10, 16, 17])) # doctest: +PROCESSOR_1_OF_3
        True
        >>> print(allequal(GNOC, [4, 5, 6, 11, 12, 13,
        ...                       18, 19, 20])) # doctest: +PROCESSOR_2_OF_3
        True

        >>> print(allequal(LNOC, [0, 1, 4, 5, 8, 9])) # doctest: +PROCESSOR_0_OF_3
        True
        >>> print(allequal(LNOC, [2, 3, 8, 9, 14, 15])) # doctest: +PROCESSOR_1_OF_3
        True
        >>> print(allequal(LNOC, [2, 3, 4, 7, 8, 9,
        ...                       12, 13, 14])) # doctest: +PROCESSOR_2_OF_3
        True


        7 cells, 2 equations, 3 processors

        ```
        0  1  2  3  4  5  6  0  1  2  3  4  5  6   cell IDs
        0  1  2  3  4  5  6  7  8  9 10 11 12 13   row IDs

        _globalOverlappingCellIDs

        0  1  2  3           0  1  2  3            proc 0
        0  1  2  3  4  5     0  1  2  3  4  5      proc  1
              2  3  4  5  6        2  3  4  5  6   proc   2

        globalOverlappingRowIDs

        0  1  2  3           7  8  9 10            proc 0
        0  1  2  3  4  5     7  8  9 10 11 12      proc  1
              2  3  4  5  6        9 10 11 12 13   proc   2

        _globalNonOverlappingCellIDs

        0  1                 0  1                  proc 0
              2  3                 2  3            proc  1
                    4  5  6              4  5  6   proc   2

        globalNonOverlappingRowIDs

        0  1                 7  8                  proc 0
              2  3                 9 10            proc  1
                    4  5  6             11 12 13   proc   2

        _localOverlappingCellIDs

        0  1  2  3           0  1  2  3            proc 0
        0  1  2  3  4  5     0  1  2  3  4  5      proc  1
              0  1  2  3  4        0  1  2  3  4   proc   2

        _localOverlappingRowIDs

        0  1  2  3           4  5  6  7            proc 0
        0  1  2  3  4  5     6  7  8  9 10 11      proc  1
              0  1  2  3  4        5  6  7  8  9   proc   2

        _localNonOverlappingCellIDs

        0  1                 0  1                  proc 0
              2  3                 2  3            proc  1
                    2  3  4              2  3  4   proc   2

        localNonOverlappingRowIDs

        0  1                 4  5                  proc 0
              2  3                 8  9            proc  1
                    2  3  4              7  8  9   proc   2
        ```
        """
        pass



from __future__ import division
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.terms.nonDiffusionTerm import _NonDiffusionTerm
from fipy.tools import inline
from fipy.tools import numerix
from fipy.terms import AbstractBaseClassError
from fipy.variables.cellVariable import CellVariable
from fipy.variables.faceVariable import FaceVariable

__all__ = ["CellTerm"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class CellTerm(_NonDiffusionTerm):
    """
    .. attention:: This class is abstract. Always create one of its subclasses.
    """
    def __init__(self, coeff=1., var=None):
        if self.__class__ is CellTerm:
            raise AbstractBaseClassError

        from fipy.variables.variable import Variable

        if not isinstance(coeff, Variable):
            from fipy.variables.constant import _Constant
            coeff = _Constant(value=coeff)

        if isinstance(coeff, FaceVariable):
             raise TypeError("The coefficient can not be a FaceVariable.")

        _NonDiffusionTerm.__init__(self, coeff=coeff, var=var)
        self.coeffVectors = None
        self._var = None

    def _checkCoeff(self, var):
        if isinstance(self.coeff, CellVariable):
            shape = self.coeff.shape[:-1]
        else:
            shape = self.coeff.shape

        if var.rank == 1 and (shape == () or len(shape) == 1):
            if len(self.coeff.shape) == 2 and isinstance(self.coeff, CellVariable):
                self.coeff *= numerix.identity(var.shape[0])[..., numerix.newaxis]
            else:
                self.coeff *= numerix.identity(var.shape[0])
            if isinstance(self.coeff, CellVariable):
                shape = self.coeff.shape[:-1]
            else:
                shape = self.coeff.shape

        if var.rank == 0:
            if shape != ():
                raise TypeError("The coefficient must be rank 0 for a rank 0 solution variable.")

        if shape != () and len(shape) != 2 and shape[0] != shape[1]:
            raise TypeError("The coefficient must be a rank-0 or rank-2 vector or a scalar value.")

        if var.rank == 1:
            if shape == ():
                pass
            elif len(shape) != 2:
                raise TypeError("The coefficient must be rank 2 or rank 0 for a rank 1 solution variable.")
            elif var.shape[0] != shape[0]:
                raise TypeError("The coefficient (N , N) shape must match the the solution variable (N,) shape.")

    def _calcCoeffVectors_(self, var, transientGeomCoeff=None, diffusionGeomCoeff=None):
        coeff = self._getGeomCoeff(var)
        weight = self._getWeight(var, transientGeomCoeff, diffusionGeomCoeff)
        if hasattr(coeff, 'old'):
            old = coeff.old
        else:
            old = coeff

        self.coeffVectors = {
            'diagonal': coeff * weight['diagonal'],
            'old value': old * weight['old value'],
            'b vector': coeff * weight['b vector'],
            'new value': coeff * weight['new value']
        }

    def _getCoeffVectors_(self, var, transientGeomCoeff=None, diffusionGeomCoeff=None):
        if self.coeffVectors is None or var is not self._var:
            self._var = var
            self._calcCoeffVectors_(var=var, transientGeomCoeff=transientGeomCoeff, diffusionGeomCoeff=diffusionGeomCoeff)

        return self.coeffVectors

    def _buildMatrixInline_(self, L, oldArray, b, dt, coeffVectors):
        oldArray = oldArray.value.ravel()
        N = len(oldArray)
        updatePyArray = numerix.zeros((N), 'd')

        dt = self._checkDt(dt)

        inline._runInline("""
            b[i] += oldArray[i] * oldCoeff[i] / dt;
            b[i] += bCoeff[i];
            updatePyArray[i] += newCoeff[i] / dt;
            updatePyArray[i] += diagCoeff[i];
        """, b=b,
            oldArray=oldArray,
            oldCoeff=coeffVectors['old value'].ravel(),
            bCoeff=coeffVectors['b vector'].ravel(),
            newCoeff=coeffVectors['new value'].ravel(),
            diagCoeff=coeffVectors['diagonal'].ravel(),
            updatePyArray=updatePyArray,
            ni=len(updatePyArray),
            dt=dt)

        L.addAtDiagonal(updatePyArray)

    def _buildMatrixNoInline_(self, L, oldArray, b, dt, coeffVectors):
        ids = self._reshapeIDs(oldArray, numerix.arange(oldArray.shape[-1]))
        b += (oldArray.value[numerix.newaxis] * coeffVectors['old value']).sum(-2).ravel() / dt
        b += coeffVectors['b vector'][numerix.newaxis].sum(-2).ravel()
        L.addAt(coeffVectors['new value'].ravel() / dt, ids.ravel(), ids.swapaxes(0, 1).ravel())
        L.addAt(coeffVectors['diagonal'].ravel(), ids.ravel(), ids.swapaxes(0, 1).ravel())

    def _buildMatrix(self, var, SparseMatrix, boundaryConditions=(), dt=None, transientGeomCoeff=None, diffusionGeomCoeff=None):

        b = numerix.zeros(var.shape, 'd').ravel()
        L = SparseMatrix(mesh=var.mesh)

        coeffVectors = self._getCoeffVectors_(var=var, transientGeomCoeff=transientGeomCoeff, diffusionGeomCoeff=diffusionGeomCoeff)

        dt = self._checkDt(dt)

        if inline.doInline and var.rank == 0:
            self._buildMatrixInline_(L=L, oldArray=var.old, b=b, dt=dt, coeffVectors=coeffVectors)
        else:
            self._buildMatrixNoInline_(L=L, oldArray=var.old, b=b, dt=dt, coeffVectors=coeffVectors)

        return (var, L, b)

    def _test(self):
        """
        The following tests demonstrate how the `CellVariable` objects
        interact with other types of `Variable` objects.

            >>> from fipy.meshes import Grid1D
            >>> from fipy.variables.cellVariable import CellVariable
            >>> from fipy.variables.faceVariable import FaceVariable
            >>> from fipy.terms.transientTerm import TransientTerm
            >>> m = Grid1D(nx=2)
            >>> cv = CellVariable(mesh=m)
            >>> fv = FaceVariable(mesh=m)
            >>> vcv = CellVariable(mesh=m, rank=1)
            >>> vfv = FaceVariable(mesh=m, rank=1)

            >>> __CellTerm(coeff=cv)
            __CellTerm(coeff=CellVariable(value=array([ 0.,  0.]), mesh=UniformGrid1D(dx=1.0, nx=2)))
            >>> __CellTerm(coeff=1)
            __CellTerm(coeff=1)
            >>> __CellTerm(coeff=fv)
            Traceback (most recent call last):
                ...
            TypeError: The coefficient can not be a FaceVariable.
            >>> TransientTerm(coeff=vcv).solve(cv)
            Traceback (most recent call last):
                ...
            TypeError: The coefficient must be rank 0 for a rank 0 solution variable.
            >>> __CellTerm(coeff=vfv)
            Traceback (most recent call last):
                ...
            TypeError: The coefficient can not be a FaceVariable.
            >>> TransientTerm(coeff=(1,)).solve(cv)
            Traceback (most recent call last):
                ...
            TypeError: The coefficient must be rank 0 for a rank 0 solution variable.

        """
        pass

class __CellTerm(CellTerm):
    """
    Dummy subclass for tests
    """
    pass


def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from PyTrilinos import ML

from fipy.solvers.trilinos.preconditioners.preconditioner import Preconditioner

__all__ = ["MultilevelDDPreconditioner"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class MultilevelDDPreconditioner(Preconditioner):
    """
    Multilevel preconditioner for Trilinos solvers. A classical smoothed
    aggregation-based 2-level domain decomposition.
    """

    def _applyToSolver(self, solver, matrix):
        if matrix.NumGlobalNonzeros() <= matrix.NumGlobalRows():
            return

        self.Prec = ML.MultiLevelPreconditioner(matrix, False)

        self.Prec.SetParameterList({text_to_native_str("output"): 0,
                                    text_to_native_str("max levels") : 2,
                                    text_to_native_str("prec type") : text_to_native_str("MGV"),
                                    text_to_native_str("increasing or decreasing") : text_to_native_str("increasing"),
                                    text_to_native_str("aggregation: type") : text_to_native_str("METIS"),
                                    text_to_native_str("aggregation: local aggregates") : 1,
                                    text_to_native_str("aggregation: damping factor") : 4. / 3.,
                                    text_to_native_str("eigen-analysis: type") : text_to_native_str("power-method"),
                                    text_to_native_str("eigen-analysis: iterations") : 20,
                                    text_to_native_str("smoother: sweeps") : 1,
                                    text_to_native_str("smoother: pre or post") : text_to_native_str("both"),
                                    text_to_native_str("smoother: type") : text_to_native_str("Aztec"),
                                    text_to_native_str("smoother: Aztec as solver") : False,
                                    text_to_native_str("coarse: type") : text_to_native_str("Amesos-KLU"),
                                    text_to_native_str("coarse: max size") : 128
                                    })

        self.Prec.ComputePreconditioner()

        solver.SetPrecOperator(self.Prec)



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.terms.abstractUpwindConvectionTerm import _AbstractUpwindConvectionTerm

__all__ = ["UpwindConvectionTerm"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class UpwindConvectionTerm(_AbstractUpwindConvectionTerm):
    r"""
    The discretization for this :class:`~fipy.terms.term.Term` is given by

    .. math::

       \int_V \nabla \cdot (\vec{u} \phi)\,dV \simeq \sum_{f} (\vec{n}
       \cdot \vec{u})_f \phi_f A_f

    where :math:`\phi_f=\alpha_f \phi_P +(1-\alpha_f)\phi_A` and
    :math:`\alpha_f` is calculated using the upwind convection scheme.
    For further details see :ref:`sec:NumericalSchemes`.
    """

    def _getDefaultSolver(self, var, solver, *args, **kwargs):
        solver = solver or super(UpwindConvectionTerm, self)._getDefaultSolver(var, solver, *args, **kwargs)
        if solver and not solver._canSolveAsymmetric():
            import warnings
            warnings.warn("%s cannot solve asymmetric matrices" % solver)
        from fipy.solvers import DefaultAsymmetricSolver
        return solver or DefaultAsymmetricSolver(*args, **kwargs)

    def _testPecletSign(self):
        r"""
            >>> from fipy import *
            >>> nx = 11
            >>> L = 1.
            >>> mesh = Grid1D(nx=nx, dx=L / nx)
            >>> var = CellVariable(mesh=mesh)

            >>> from fipy.variables.faceVariable import FaceVariable
            >>> convCoeff = FaceVariable(mesh=mesh, rank=1)
            >>> diffCoeff = FaceVariable(mesh=mesh, value=1e-20)
            >>> x = mesh.faceCenters[0]
            >>> D = 1000.
            >>> u = -1e+6
            >>> convCoeff.setValue((u,), where=x < L / 2)
            >>> diffCoeff.setValue(D, where=x > L / 2)

            >>> var.constrain(1., mesh.facesRight)
            >>> var.constrain(10, mesh.facesLeft)

            >>> dTerm = DiffusionTerm(diffCoeff)

            >>> x = mesh.cellCenters[0]
            >>> var0 = 2 * D / (-u * L + 2 * D)
            >>> analytical = 2 * (1 - var0) * x / L + 2 * var0 - 1
            >>> analytical = var0 * (x < L / 2) + analytical * (x >= L / 2)

            >>> var[:] = 0
            >>> eqn = UpwindConvectionTerm(coeff=convCoeff) == dTerm
            >>> eqn.solve(var)
            >>> print(var.allclose(analytical))
            1

            >>> var[:] = 0
            >>> eqn = TransientTerm(1e-10) == UpwindConvectionTerm(coeff=-convCoeff) +  dTerm
            >>> eqn.solve(var, dt = 1e+10)
            >>> print(var.allclose(analytical))
            1

            >>> var[:] = 0
            >>> eqn = 0 == UpwindConvectionTerm(coeff=-convCoeff) +  dTerm
            >>> eqn.solve(var)
            >>> print(var.allclose(analytical))
            1

            >>> var[:] = 0
            >>> eqn = 0 == -UpwindConvectionTerm(coeff=convCoeff) +  dTerm
            >>> eqn.solve(var)
            >>> print(var.allclose(analytical))
            1

        """
        pass

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()




from __future__ import unicode_literals
from fipy.solvers.pyamgx import PyAMGXSolver
from fipy.solvers.pyamgx.preconditioners import *

__all__ = ["LinearBiCGStabSolver"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class LinearBiCGStabSolver(PyAMGXSolver):
    """
    The `LinearBiCGStabSolver` is an interface to the PBICGSTAB solver in
    AMGX, with a Jacobi preconditioner by default.
    """

    def __init__(self, tolerance=1e-10, iterations=2000,
            precon=BlockJacobiPreconditioner(),
                 **kwargs):
        """
        Parameters
        ----------
        tolerance : float
            Required error tolerance.
        iterations : int
            Maximum number of iterative steps to perform.
        precon : ~fipy.solvers.pyamgx.preconditioners.preconditioners.Preconditioner, optional
        **kwargs
            Other AMGX solver options
        """
        config_dict = {
            "config_version": 2,
            "determinism_flag": 1,
            "exception_handling" : 1,
            "solver": {
                "convergence": "RELATIVE_INI_CORE",
                "monitor_residual": 1,
                "solver": "PBICGSTAB",
                "preconditioner": {
                    "solver": "NOSOLVER"
                }
            }
        }
        super(LinearBiCGStabSolver, self).__init__(
                config_dict,
                tolerance=tolerance,
                iterations=iterations,
                precon=precon,
                **kwargs)



from __future__ import unicode_literals
from fipy.solvers.scipy.linearLUSolver import LinearLUSolver

__all__ = ["LinearLUSolver"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.terms.abstractConvectionTerm import _AbstractConvectionTerm


class _AsymmetricConvectionTerm(_AbstractConvectionTerm):

    def _getDefaultSolver(self, var, solver, *args, **kwargs):
        r"""
        Make sure the method actually does something.
        >>> print(_AsymmetricConvectionTerm((1,)).getDefaultSolver().__repr__()[:6])
        Linear
        """
        solver = solver or super(_AsymmetricConvectionTerm, self)._getDefaultSolver(var, solver, *args, **kwargs)
        if solver and not solver._canSolveAsymmetric():
            import warnings
            warnings.warn("%s cannot solve asymmetric matrices" % solver)
        from fipy.solvers import DefaultAsymmetricSolver
        return solver or DefaultAsymmetricSolver(*args, **kwargs)

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()




from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.terms.abstractConvectionTerm import _AbstractConvectionTerm
from fipy.variables.faceVariable import FaceVariable
from fipy.tools.dimensions.physicalField import PhysicalField
from fipy.tools import inline
from fipy.tools import numerix

class _UpwindConvectionTermAlpha(FaceVariable):
    def __init__(self, P):
        FaceVariable.__init__(self, mesh=P.mesh, elementshape=P.shape[:-1])
        self.P = self._requires(P)

    if inline.doInline:
        def _calcValue(self):
            P  = self.P.numericValue
            alpha = self._array.copy()

            inline._runInline("""
                alpha[i] = 0.5;

                if (P[i] > 0.) {
                    alpha[i] = 1.;
                } else {
                    alpha[i] = 0.;
                }
            """,
            alpha=alpha, P=P,
            ni = len(P.flat))

            return self._makeValue(value=alpha)
    else:
        def _calcValue(self):
            P  = self.P.numericValue
            alpha = numerix.where(P > 0., 1., 0.)
            return PhysicalField(value=alpha)

class _AbstractUpwindConvectionTerm(_AbstractConvectionTerm):
    def _alpha(self, P):
        return _UpwindConvectionTermAlpha(P)

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.terms.explicitSourceTerm import _ExplicitSourceTerm
from fipy.variables.cellVariable import CellVariable

__all__ = ["ResidualTerm"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class ResidualTerm(_ExplicitSourceTerm):
    r"""

    The `ResidualTerm` is a special form of explicit `SourceTerm` that adds the
    residual of one equation to another equation. Useful for Newton's method.
    """
    def __init__(self, equation, underRelaxation=1.):
        self.equation = equation
        self.underRelaxation = underRelaxation

        _ExplicitSourceTerm.__init__(self, var=None)

    def __repr__(self):
        return r"$\Delta$[" + repr(self.equation) + "]"

    def _getGeomCoeff(self, var):
        return self.coeff

    def _buildMatrix(self, var, SparseMatrix, boundaryConditions=(), dt=None, transientGeomCoeff=None, diffusionGeomCoeff=None):
        vec = self.equation.justResidualVector(var=None,
                                               boundaryConditions=boundaryConditions,
                                               dt=dt)

        self.coeff = CellVariable(mesh=var.mesh, value=vec * self.underRelaxation)
        self.geomCoeff = None
        self.coeffVectors = None

        return _ExplicitSourceTerm._buildMatrix(self, var=var, SparseMatrix=SparseMatrix, boundaryConditions=boundaryConditions, dt=dt, transientGeomCoeff=transientGeomCoeff, diffusionGeomCoeff=diffusionGeomCoeff)



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.terms.cellTerm import CellTerm
from fipy.terms import AbstractBaseClassError
from fipy.variables.cellVariable import CellVariable
from fipy.tools import numerix

__all__ = ["SourceTerm"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class SourceTerm(CellTerm):
    """
    .. attention:: This class is abstract. Always create one of its subclasses.
    """
    def __init__(self, coeff=0., var=None):
        if self.__class__ is SourceTerm:
            raise AbstractBaseClassError
        CellTerm.__init__(self, coeff=coeff, var=var)

    def _calcGeomCoeff(self, var):
        self._checkCoeff(var)

        if self.coeff.shape != () and self.coeff.shape[-1] != len(var.mesh.cellVolumes):
            return self.coeff[..., numerix.newaxis] * CellVariable(mesh=var.mesh, value=var.mesh.cellVolumes)
        else:
            return self.coeff * CellVariable(mesh=var.mesh, value=var.mesh.cellVolumes)

    def _checkDt(self, dt):
        return 1.



"""This module provides a generic, high-level interface for
creating shared temporary files.  All of the interfaces
provided by this module can be used without fear of race conditions.
"""

from tempfile import NamedTemporaryFile, template, _TemporaryFileWrapper
from future import utils

from fipy.tools import parallelComm

__all__ = ["SharedTemporaryFile"]

# Silence linting errors that come from following conventions of
# tempfile package.
# pylint: disable=invalid-name
# pylint: disable=redefined-builtin
# pylint: disable=too-many-arguments

if utils.PY3:
    import io as _io

    def SharedTemporaryFile(
        mode="w+b",
        buffering=-1,
        encoding=None,
        newline=None,
        suffix="",
        prefix=template,
        dir=None,
        delete=True,
        communicator=parallelComm,
    ):
        """Create a temporary file shared by all MPI ranks.

        The file is created as `NamedTemporaryFile` would do it.
        The name of the returned file-like object is accessible as its
        ``name`` attribute.  The file will be automatically deleted when it
        is closed unless the `delete` argument is set to False.

        >>> from fipy.tools import SharedTemporaryFile, parallelComm
        >>> with SharedTemporaryFile(mode='w+', suffix=".tmp") as tmpFile:
        ...     # write on processor 0
        ...     if parallelComm.procID == 0:
        ...         _ = tmpFile.write("shared text")
        ...
        ...     parallelComm.Barrier()
        ...
        ...     # read on all processors
        ...     _ = tmpFile.seek(0)
        ...     txt = tmpFile.read()
        >>> print(txt)
        shared text

        Parameters
        ----------
        prefix, suffix, dir : str
            As for mkstemp
        mode : str
            The mode argument to io.open (default "w+b")
        buffering : int
            The buffer size argument to io.open (default -1)
        encoding : str or None
            The encoding argument to io.open (default None)
        newline : str or None
            The newline argument to io.open (default None)
        delete : bool
            Whether the file is deleted on close (default True)
        communicator : ~fipy.tools.comms.commWrapper.CommWrapper
            MPI communicator describing ranks to share with.  A duck-typed
            object with `procID` and `Nproc` attributes is sufficient.

        Returns
        -------
        file-like object

        See Also
        --------
        tempfile.NamedTemporaryFile, tempfile.mkstemp, io.open

        """

        if communicator.procID == 0:
            file = NamedTemporaryFile(
                mode=mode,
                buffering=buffering,
                encoding=encoding,
                newline=newline,
                suffix=suffix,
                prefix=prefix,
                dir=dir,
                delete=delete,
            )
            fname = file.name
        else:
            fname = None

        fname = communicator.bcast(fname)

        if communicator.procID != 0:
            file = _io.open(
                fname, mode, buffering=buffering, newline=newline, encoding=encoding
            )
            # let procID 0 handle delete
            file = _TemporaryFileWrapper(file, fname, delete=False)

        return file


else:

    def SharedTemporaryFile(
        mode="w+b",
        bufsize=-1,
        suffix="",
        prefix=template,
        dir=None,
        delete=True,
        communicator=parallelComm,
    ):
        """Create a temporary file shared by all MPI ranks.

        The file is created as `NamedTemporaryFile` would do it.
        The name of the returned file-like object is accessible as its
        ``name`` attribute.  The file will be automatically deleted when it
        is closed unless the `delete` argument is set to False.

        >>> from fipy.tools import SharedTemporaryFile, parallelComm
        >>> with SharedTemporaryFile(mode='w+', suffix=".tmp") as tmpFile:
        ...     # write on processor 0
        ...     if parallelComm.procID == 0:
        ...         _ = tmpFile.write("shared text")
        ...
        ...     parallelComm.Barrier()
        ...
        ...     # read on all processors
        ...     _ = tmpFile.seek(0)
        ...     txt = tmpFile.read()
        >>> print(txt)
        shared text

        Parameters
        ----------
        prefix, suffix, dir : str
            As for mkstemp
        mode : str
            The mode argument to os.fdopen (default "w+b")
        bufsize : int
            The buffer size argument to os.fdopen (default -1)
        delete : bool
            Whether the file is deleted on close (default True)
        communicator : ~fipy.tools.comms.commWrapper.CommWrapper
            MPI communicator describing ranks to share with.  A duck-typed
            object with `procID` and `Nproc` attributes is sufficient

        Returns
        -------
        file-like object

        See Also
        --------
        tempfile.NamedTemporaryFile, tempfile.mkstemp, os.fdopen

        """

        if communicator.procID == 0:
            file = NamedTemporaryFile(
                mode=mode,
                bufsize=bufsize,
                suffix=suffix,
                prefix=prefix,
                dir=dir,
                delete=delete,
            )
            fname = file.name
        else:
            fname = None

        fname = communicator.bcast(fname)

        if communicator.procID != 0:
            file = open(fname, mode, bufsize)
            # let procID 0 handle delete
            file = _TemporaryFileWrapper(file, fname, delete=False)

        return file


def _test():
    import fipy.tests.doctestPlus

    return fipy.tests.doctestPlus.testmod()


if __name__ == "__main__":
    _test()



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.meshes.abstractMesh import AbstractMesh

__all__ = ["UniformGrid"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class UniformGrid(AbstractMesh):
    """Wrapped scaled geometry properties"""
    @property
    def _scaledFaceAreas(self):
        return self._faceAreas

    @property
    def _scaledCellVolumes(self):
        return self._cellVolumes

    @property
    def _scaledCellCenters(self):
        return self._cellCenters

    @property
    def _scaledCellDistances(self):
        return self._cellDistances

    @property
    def _scaledCellToCellDistances(self):
        return self._cellToCellDistances

    @property
    def _scaledFaceToCellDistances(self):
        return self._faceToCellDistances

    """Geometry properties common to 1D, 2D, 3D"""
    @property
    def _orientedFaceNormals(self):
        return self.faceNormals

    @property
    def _faceCellToCellNormals(self):
        return self.faceNormals

    def _getFaceToCellDistances(self):
        return self._internalFaceToCellDistances

    def _setFaceToCellDistances(self, v):
        self._internalFaceToCellDistances = v
        self._setScaledValues()

    _faceToCellDistances = property(_getFaceToCellDistances,
                                    _setFaceToCellDistances)



from __future__ import unicode_literals
from builtins import object
__docformat__ = 'restructuredtext'

__all__ = []

class _AbstractRepresentation(object):
    def __init__(self, mesh):
        self.mesh = mesh

    def getstate(self):
        """Collect the necessary information to ``pickle`` the `Mesh` to persistent storage.
        """
        raise NotImplemented

    def setstate(self, state):
        """Populate a new `Mesh` from ``pickled`` persistent storage.
        """
        raise NotImplemented

    def repr(self):
        raise NotImplemented



from __future__ import unicode_literals
from builtins import object
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.tools import numerix

class _CoupledCellVariable(object):
    def __init__(self, vars):
        self.vars = vars

    @property
    def shape(self):
        return (len(self.vars) * self.mesh.numberOfCells,)

    @property
    def rank(self):
        return self.vars[0].rank

    def __repr__(self):
        return "(" + ", ".join([repr(var) for var in self.vars]) + ")"

    @property
    def mesh(self):
        meshes = list(set([var.mesh for var in self.vars]))

        if len(meshes) == 0:
            raise Exception("There are no Meshes defined")
        elif len(meshes) > 1:
            raise Exception("All Variables must be defined on the same Mesh")
        else:
            return meshes[0]

    def __getitem__(self, index):
        return numerix.concatenate([numerix.array(var[index]) for var in self.vars])

    def __setitem__(self, index, value):
        N = self.mesh.numberOfCells
        for i, var in enumerate(self.vars):
            if numerix.shape(value) == ():
                var[index] = value
            else:
                var[index] = value[i * N:(i + 1) * N]

    def _getValue(self):
        return numerix.concatenate([numerix.array(var.value) for var in self.vars])

    def _setValue(self, value):
        self[:] = value

    value = property(_getValue, _setValue)

    @property
    def globalValue(self):
        return numerix.concatenate([numerix.array(var.globalValue) for var in self.vars])

    @property
    def numericValue(self):
        return numerix.concatenate([var.numericValue for var in self.vars])

    @property
    def unit(self):
        from fipy.tools.dimensions import physicalField
        return physicalField._unity

    def __array__(self, t=None):
        """
        Attempt to convert the `_CoupledCellVariable` to a numerix `array` object

        >>> from fipy import *
        >>> mesh = Grid1D(nx=2)
        >>> v1 = CellVariable(mesh=mesh, value=[2, 3])
        >>> v2 = CellVariable(mesh=mesh, value=[4, 5])
        >>> v = _CoupledCellVariable(vars=(v1, v2))
        >>> print(numerix.allequal([2, 3, 4, 5], numerix.array(v))) # doctest: +PROCESSOR_0
        True
        >>> v[:] = (6, 7, 8, 9)
        >>> print(v1)
        [6 7]
        >>> print(v2)
        [8 9]
        >>> v.getsctype() == numerix.NUMERIX.obj2sctype(numerix.array(1))
        True

        """
        return numerix.array(self.value, t)

    def __neg__(self):
        return _CoupledCellVariable([-var for var in self.vars])

    def __abs__(self):
        return _CoupledCellVariable([abs(var) for var in self.vars])

    def __iter__(self):
        return iter(self.value)

    def getsctype(self, default=None):
        if not hasattr(self, 'typecode'):
            self.typecode = numerix.obj2sctype(rep=self.numericValue, default=default)
        return self.typecode

    def ravel(self):
        return self.value.ravel()

    def copy(self):
        return self.__class__(vars=[var.copy() for var in self.vars])

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import unicode_literals
from builtins import str
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.meshes.representations.abstractRepresentation import _AbstractRepresentation

class _GridRepresentation(_AbstractRepresentation):

    def getstate(self):
        """Collect the necessary information to ``pickle`` the `Grid` to persistent storage.
        """
        args = self.mesh.args.copy()
        args["_RepresentationClass"] = self.__class__
        return args

    @staticmethod
    def setstate(mesh, state):
        """Create a new `Grid` from ``pickled`` persistent storage.
        """
        mesh.__init__(**state)

    def _repr(self, dns):
        dnstr = []
        for d, n in dns:
            dnstr.append(d + "=" + str(self.mesh.args[d]))
            if self.mesh.args[n] is not None:
                dnstr.append(n + "=" + str(self.mesh.args[n]))

        return "%s(%s)" % (self.mesh.__class__.__name__, ", ".join(dnstr))

    def _test(self):
        """

        Check that the following grid classes can be pickled and unpickled.

        >>> import fipy as fp

        >>> m = fp.PeriodicGrid2DLeftRight(nx=10, ny=10)
        >>> v = fp.CellVariable(mesh=m, value=m.x)
        >>> (f, filename) = fp.dump.write(v, extension='.gz')
        >>> v0 = fp.dump.read(filename, f)
        >>> print((v == v0.mesh.x).all())
        True

        >>> m = fp.PeriodicGrid1D(nx=10)
        >>> v = fp.CellVariable(mesh=m, value=m.x)
        >>> (f, filename) = fp.dump.write(v, extension='.gz')
        >>> v0 = fp.dump.read(filename, f)
        >>> print((v == v0.mesh.x).all())
        True

        >>> m = fp.Tri2D(nx=10, ny=10)
        >>> v = fp.CellVariable(mesh=m, value=m.x)
        >>> (f, filename) = fp.dump.write(v, extension='.gz')
        >>> v0 = fp.dump.read(filename, f)
        >>> print((v == v0.mesh.x).all())
        True

        >>> m = fp.SkewedGrid2D(nx=10, ny=10)
        >>> v = fp.CellVariable(mesh=m, value=m.x)
        >>> (f, filename) = fp.dump.write(v, extension='.gz')
        >>> v0 = fp.dump.read(filename, f)
        >>> print((v == v0.mesh.x).all())
        True

        """

class _Grid1DRepresentation(_GridRepresentation):

    def repr(self):
        return self._repr(dns=[("dx", "nx")])

class _Grid2DRepresentation(_GridRepresentation):

    def repr(self):
        return self._repr(dns=[("dx", "nx"), ("dy", "ny")])

class _Grid3DRepresentation(_GridRepresentation):

    def repr(self):
        return self._repr(dns=[("dx", "nx"), ("dy", "ny"), ("dz", "nz")])

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()




from __future__ import unicode_literals
__all__ = []

from fipy.tests.doctestPlus import _LateImportDocTestSuite
import fipy.tests.testProgram

def _suite():

    return _LateImportDocTestSuite(docTestModuleNames = (
            'cellTerm',
            'abstractDiffusionTerm',
            'diffusionTerm',
            'term',
            'abstractConvectionTerm',
            'transientTerm',
            'powerLawConvectionTerm',
            'exponentialConvectionTerm',
            'upwindConvectionTerm',
            'implicitSourceTerm',
            'coupledBinaryTerm',
            'abstractBinaryTerm',
            'unaryTerm',
            'nonDiffusionTerm',
            'asymmetricConvectionTerm',
            'binaryTerm',
            'firstOrderAdvectionTerm',
            'advectionTerm',
            'vanLeerConvectionTerm'
            ), base = __name__)

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.variables.cellVariable import CellVariable
from fipy.tools.numerix import MA
from fipy.tools import numerix

__all__ = []

class _InterfaceFlagVariable(CellVariable):
    def __init__(self, distanceVar):
        """
        Creates an `_InterfaceFlagVariable` object.

        Parameters
        ----------
        distanceVar : ~fipy.variables.distanceVariable.DistanceVariable
        """
        CellVariable.__init__(self, distanceVar.mesh, hasOld=False)
        self.distanceVar = self._requires(distanceVar)

    def _calcValue(self):
        flag = MA.filled(numerix.take(self.distanceVar._interfaceFlag, self.mesh.cellFaceIDs), 0)
        flag = numerix.sum(flag, axis=0)
        return numerix.where(numerix.logical_and(self.distanceVar.value > 0, flag > 0), 1, 0)



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.terms.abstractDiffusionTerm import _AbstractDiffusionTerm

__all__ = ["DiffusionTermNoCorrection"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class DiffusionTermNoCorrection(_AbstractDiffusionTerm):
    def _getNormals(self, mesh):
        return mesh.faceNormals

    def _treatMeshAsOrthogonal(self, mesh):
        return True



from __future__ import unicode_literals
from builtins import str
import os
from importlib import import_module

from fipy.tools.parser import _parseSolver

from fipy.solvers.solver import *
__all__ = list(solver.__all__)
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

_desired_solver = _parseSolver()

if _desired_solver is None and 'FIPY_SOLVERS' in os.environ:
    _desired_solver = os.environ['FIPY_SOLVERS'].lower()
del os
    
try:
    from mpi4py import MPI
    _Nproc = MPI.COMM_WORLD.size
    del MPI
except ImportError:
    _Nproc = 1

_exceptions = []

class SerialSolverError(Exception):
    def __init__(self):
        super(SerialSolverError, self).__init__('solver does not run in parallel')

_exceptions = {}

def _import_matrix(suite, kind):
    """`from fipy.matrices.suiteMatrix import _SuiteKindMatrix`
    """
    m = import_module("fipy.matrices.{}Matrix".format(suite.lower()))
    return getattr(m, "_{}{}Matrix".format(suite, kind))

def _import_mesh_matrices(suite):
    """Import row, column, and general mesh matrices from `suite`
    """
    _RowMeshMatrix = _import_matrix(suite=suite, kind="RowMesh")
    _ColMeshMatrix = _import_matrix(suite=suite, kind="ColMesh")
    _MeshMatrix = _import_matrix(suite=suite, kind="Mesh")

    return _RowMeshMatrix, _ColMeshMatrix, _MeshMatrix

solver = None

from fipy.tools.comms.dummyComm import DummyComm
serialComm, parallelComm = DummyComm(), DummyComm()

if solver is None and _desired_solver in ["pysparse", None]:
    try:
        if _Nproc > 1:
            raise SerialSolverError()
        from fipy.solvers.pysparse import *
        __all__.extend(pysparse.__all__)
        _mesh_matrices = _import_mesh_matrices(suite="Pysparse")
        solver = "pysparse"
    except Exception as inst:
        _exceptions["pysparse"] = inst

if solver is None and _desired_solver in ["petsc", None]:
    try:
        import petsc4py
        petsc4py.init()

        from fipy.solvers.petsc import *
        __all__.extend(petsc.__all__)

        from fipy.solvers.petsc.comms.serialPETScCommWrapper import SerialPETScCommWrapper
        serialComm = SerialPETScCommWrapper()

        if _Nproc > 1:
            from fipy.solvers.petsc.comms.parallelPETScCommWrapper import ParallelPETScCommWrapper
            parallelComm = ParallelPETScCommWrapper()
        else:
            parallelComm = SerialPETScCommWrapper()

        _mesh_matrices = _import_mesh_matrices(suite="PETSc")
        solver = "petsc"
    except Exception as inst:
        _exceptions["petsc"] = inst

if solver is None and _desired_solver in ["trilinos", "no-pysparse", None]:
    try:
        from fipy.solvers.trilinos import *
        __all__.extend(trilinos.__all__)
        
        from fipy.solvers.trilinos.comms.serialEpetraCommWrapper import SerialEpetraCommWrapper
        serialComm = SerialEpetraCommWrapper()

        if _Nproc > 1:
            from fipy.solvers.trilinos.comms.parallelEpetraCommWrapper import ParallelEpetraCommWrapper
            parallelComm = ParallelEpetraCommWrapper()
        else:
            parallelComm = SerialEpetraCommWrapper()

        if _desired_solver != "no-pysparse":
            try:
                _mesh_matrices = _import_mesh_matrices(suite="Pysparse")
                solver = "trilinos"
            except ImportError:
                pass
                
        if solver is None:
            # no-pysparse requested or pysparseMatrix failed to import
            _mesh_matrices = _import_mesh_matrices(suite="Trilinos")
            solver = "no-pysparse"
    except Exception as inst:
        _exceptions["trilinos"] = inst

if solver is None and _desired_solver in ["scipy", None]:
    try:
        if _Nproc > 1:
            raise SerialSolverError()
        from fipy.solvers.scipy import *
        __all__.extend(scipy.__all__)
        _mesh_matrices = _import_mesh_matrices(suite="Scipy")
        solver = "scipy"
    except Exception as inst:
        _exceptions["scipy"] = inst

if solver is None and _desired_solver in ["pyamg", None]:
    try:
        if _Nproc > 1:
            raise SerialSolverError()
        from fipy.solvers.pyAMG import *
        __all__.extend(pyAMG.__all__)
        _mesh_matrices = _import_mesh_matrices(suite="Scipy")
        solver = "pyamg"
    except Exception as inst:
        _exceptions["pyamg"] = inst

if solver is None and _desired_solver in ["pyamgx", None]:
    try:
        if _parallelComm.Nproc > 1:
            raise  SerialSolverError('pyamgx')
        from fipy.solvers.pyamgx import *
        __all__.extend(pyamgx.__all__)
        _mesh_matrices = _import_mesh_matrices(suite="Scipy")
        solver = "pyamgx"
    except Exception as inst:
        _exceptions["pyamgx"] = inst

if solver is None:
    if _desired_solver is None:
        raise ImportError('Unable to load a solver: %s' % str(_exceptions))
    else:
        if len(_exceptions) > 0:
            raise ImportError('Unable to load solver %s: %s' % (_desired_solver, _exceptions[_desired_solver]))
        else:
            raise ImportError('Unknown solver package %s' % _desired_solver)

# don't unpack until here in order to keep code above more succinct
_RowMeshMatrix, _ColMeshMatrix, _MeshMatrix = _mesh_matrices

from fipy.tests.doctestPlus import register_skipper

register_skipper(flag='PYSPARSE_SOLVER',
                 test=lambda: solver == 'pysparse',
                 why="the Pysparse solvers are not being used.",
                 skipWarning=True)

register_skipper(flag='NOT_PYAMGX_SOLVER',
                 test=lambda: solver != 'pyamgx',
                 why="the PyAMGX solver is being used.",
                 skipWarning=True)
del register_skipper



"""
2D cylindrical rectangular Mesh with constant spacing in x and constant spacing in y
"""
from __future__ import division
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.meshes.uniformGrid2D import UniformGrid2D
from fipy.tools import numerix
from fipy.tools import parallelComm

__all__ = ["CylindricalUniformGrid2D"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class CylindricalUniformGrid2D(UniformGrid2D):
    r"""
    Creates a 2D cylindrical grid in the radial and axial directions,
    appropriate for axial symmetry.
    """
    def __init__(self, dx=1., dy=1., nx=1, ny=1, origin=((0,), (0,)),
                 overlap=2, communicator=parallelComm, *args, **kwargs):
        super(CylindricalUniformGrid2D, self).__init__(dx=dx, dy=dy, nx=nx, ny=ny,
                                                       origin=origin, overlap=overlap,
                                                       communicator=communicator,
                                                       *args, **kwargs)

    def _translate(self, vector):
        return CylindricalUniformGrid2D(dx = self.args['dx'], nx = self.args['nx'],
                                        dy = self.args['dy'], ny = self.args['ny'],
                                        origin=numerix.array(self.args['origin']) + vector,
                                        overlap=self.args['overlap'])

    @property
    def _faceAreas(self):
        faceAreas = numerix.zeros(self.numberOfFaces, 'd')
        faceAreas[:self.numberOfHorizontalFaces] = self.dx
        faceAreas[self.numberOfHorizontalFaces:] = self.dy
        return faceAreas * self._faceCenters[0]

    @property
    def _cellVolumes(self):
        return numerix.ones(self.numberOfCells, 'd') * self.dx * self.dy

    @property
    def _cellAreas(self):
        areas = numerix.ones((4, self.numberOfCells), 'd')
        areas[0] = self.dx * self._cellCenters[0]
        areas[1] = self.dy * (self._cellCenters[0] + self.dx / 2)
        areas[2] = self.dx * self._cellCenters[0]
        areas[3] = self.dy * (self._cellCenters[0] - self.dx / 2)
        return areas

#     def _calcAreaProjections(self):
#         return self._getAreaProjectionsPy()

    def _calcAreaProjections(self):
        return self.faceNormals * self._faceAreas

    @property
    def cellVolumes(self):
        return self._cellVolumes * self.cellCenters[0].value

    def _test(self):
        """
        These tests are not useful as documentation, but are here to ensure
        everything works as expected.

            >>> import fipy as fp

            >>> dx = 0.5
            >>> dy = 2.
            >>> nx = 3
            >>> ny = 2

            >>> mesh = CylindricalUniformGrid2D(nx = nx, ny = ny, dx = dx, dy = dy)

            >>> vertices = numerix.array(((0., 1., 2., 3., 0., 1.,
            ...                            2., 3., 0., 1., 2., 3.),
            ...                           (0., 0., 0., 0., 1., 1.,
            ...                            1., 1., 2., 2., 2., 2.)))
            >>> vertices *= numerix.array([[dx], [dy]])
            >>> print(numerix.allequal(vertices,
            ...                        mesh.vertexCoords)) # doctest: +PROCESSOR_0
            True

            >>> faces = numerix.array([[0, 1, 2, 4, 5, 6, 8, 9, 10,
            ...                         0, 1, 2, 3, 4, 5, 6, 7],
            ...                        [1, 2, 3, 5, 6, 7, 9, 10, 11,
            ...                         4, 5, 6, 7, 8, 9, 10, 11]])
            >>> print(numerix.allequal(faces,
            ...                        mesh.faceVertexIDs)) # doctest: +PROCESSOR_0
            True

            >>> cells = numerix.array(((0,  1,  2,  3,  4,  5),
            ...                       (10, 11, 12, 14, 15, 16),
            ...                       ( 3,  4,  5,  6,  7,  8),
            ...                       ( 9, 10, 11, 13, 14, 15)))
            >>> print(numerix.allequal(cells,
            ...                        mesh.cellFaceIDs)) # doctest: +PROCESSOR_0
            True

            >>> externalFaces = numerix.array((0, 1, 2, 6, 7, 8, 9, 12, 13, 16))
            >>> print(numerix.allequal(externalFaces,
            ...                        numerix.nonzero(mesh.exteriorFaces))) # doctest: +PROCESSOR_0
            True

            >>> internalFaces = numerix.array((3, 4, 5, 10, 11, 14, 15))
            >>> print(numerix.allequal(internalFaces,
            ...                        numerix.nonzero(mesh.interiorFaces))) # doctest: +PROCESSOR_0
            True

            >>> from fipy.tools.numerix import MA
            >>> faceCellIds = MA.masked_values((( 0,  1,  2, 0,  1,  2,  3,  4,
            ...                                   5,  0,  0, 1,  2,  3,  3,  4,  5),
            ...                                 (-1, -1, -1, 3,  4,  5, -1, -1,
            ...                                  -1, -1,  1, 2, -1, -1,  4,  5, -1)), -1)
            >>> print(numerix.allequal(faceCellIds, mesh.faceCellIDs)) # doctest: +PROCESSOR_0
            True

            >>> faceCoords = numerix.take(vertices, faces, axis=1)
            >>> faceCenters = (faceCoords[..., 0,:] + faceCoords[..., 1,:]) / 2.
            >>> print(numerix.allclose(faceCenters, mesh.faceCenters, atol = 1e-10, rtol = 1e-10))
            True

            >>> faceAreas = numerix.array((dx, dx, dx, dx, dx, dx, dx, dx, dx,
            ...                            dy, dy, dy, dy, dy, dy, dy, dy))
            >>> faceAreas = faceAreas * faceCenters[0]
            >>> print(numerix.allclose(faceAreas, mesh._faceAreas, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> faceNormals = numerix.array(((0., 0., 0., 0., 0., 0., 0., 0., 0.,
            ...                               -1., 1., 1., 1., -1., 1., 1., 1.),
            ...                              (-1., -1., -1., 1., 1., 1., 1., 1.,
            ...                               1., 0, 0, 0, 0, 0, 0, 0, 0)))
            >>> print(numerix.allclose(faceNormals, mesh.faceNormals, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cellToFaceOrientations = numerix.array(((1,  1,  1, -1, -1, -1),
            ...                                         (1,  1,  1,  1,  1,  1),
            ...                                         (1,  1,  1,  1,  1,  1),
            ...                                         (1, -1, -1,  1, -1, -1)))
            >>> print(numerix.allequal(cellToFaceOrientations, mesh._cellToFaceOrientations)) # doctest: +PROCESSOR_0
            True

            >>> type(mesh.cellCenters)
            <class 'fipy.variables.cellVariable.CellVariable'>

            >>> testCellVolumes = mesh.cellCenters[0].globalValue * numerix.array((dx*dy, dx*dy, dx*dy, dx*dy, dx*dy, dx*dy))

            >>> print(isinstance(mesh.cellVolumes, numerix.ndarray))
            True

            >>> globalValue = fp.CellVariable(mesh=mesh, value=mesh.cellVolumes).globalValue
            >>> print(numerix.allclose(testCellVolumes, globalValue, atol = 1e-10, rtol = 1e-10))
            True

            >>> cellCenters = numerix.array(((dx/2., 3.*dx/2., 5.*dx/2.,    dx/2., 3.*dx/2., 5.*dx/2.),
            ...                              (dy/2.,    dy/2.,    dy/2., 3.*dy/2., 3.*dy/2., 3.*dy/2.)))
            >>> print(numerix.allclose(cellCenters, mesh.cellCenters, atol = 1e-10, rtol = 1e-10))
            True

            >>> cellDistances = numerix.array((dy / 2., dy / 2., dy / 2.,
            ...                                dy, dy, dy,
            ...                                dy / 2., dy / 2., dy / 2.,
            ...                                dx / 2., dx, dx,
            ...                                dx / 2.,
            ...                                dx / 2., dx, dx,
            ...                                dx / 2.))
            >>> print(numerix.allclose(cellDistances, mesh._cellDistances, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> faceToCellDistances = MA.masked_values(((dy / 2., dy / 2., dy / 2., dy / 2., dy / 2., dy / 2., dy / 2., dy / 2., dy / 2., dx / 2., dx / 2., dx / 2., dx / 2., dx / 2., dx / 2., dx / 2., dx / 2.),
            ...                                         (     -1,      -1,      -1, dy / 2., dy / 2., dy / 2.,      -1,      -1,      -1,      -1, dx / 2., dx / 2.,      -1,      -1, dx / 2., dx / 2.,      -1)), -1)
            >>> faceToCellDistanceRatios = faceToCellDistances[0] / cellDistances
            >>> print(numerix.allclose(faceToCellDistanceRatios, mesh._faceToCellDistanceRatio, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> areaProjections = faceNormals * faceAreas
            >>> print(numerix.allclose(areaProjections, mesh._areaProjections, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> tangents1 = numerix.array(((1., 1., 1., -1., -1., -1., -1., -1.,
            ...                             -1., 0., 0., 0., 0., 0., 0., 0., 0.),
            ...                            (0, 0, 0, 0, 0, 0, 0, 0, 0, -1., 1.,
            ...                             1., 1., -1., 1., 1., 1.)))
            >>> print(numerix.allclose(tangents1, mesh._faceTangents1, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> tangents2 = numerix.zeros((2, 17), 'd')
            >>> print(numerix.allclose(tangents2, mesh._faceTangents2, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cellToCellIDs = MA.masked_values(((-1, -1, -1,  0,  1,  2),
            ...                                   ( 1,  2, -1,  4,  5, -1),
            ...                                   ( 3,  4,  5, -1, -1, -1),
            ...                                   (-1,  0,  1, -1,  3,  4)), -1)
            >>> print(numerix.allequal(cellToCellIDs, mesh._cellToCellIDs)) # doctest: +PROCESSOR_0
            True

            >>> cellToCellDistances = MA.masked_values(((dy / 2., dy / 2., dy / 2.,      dy,      dy,      dy),
            ...                                         (     dx,      dx, dx / 2.,      dx,      dx, dx / 2.),
            ...                                         (     dy,      dy,      dy, dy / 2., dy / 2., dy / 2.),
            ...                                         (dx / 2.,      dx,      dx, dx / 2.,      dx,      dx)), -1)
            >>> print(numerix.allclose(cellToCellDistances, mesh._cellToCellDistances, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cellNormals = numerix.array(((( 0,  0,  0,  0,  0,  0),
            ...                               ( 1,  1,  1,  1,  1,  1),
            ...                               ( 0,  0,  0,  0,  0,  0),
            ...                               (-1, -1, -1, -1, -1, -1)),
            ...                              ((-1, -1, -1, -1, -1, -1),
            ...                               ( 0,  0,  0,  0,  0,  0),
            ...                               ( 1,  1,  1,  1,  1,  1),
            ...                               ( 0,  0,  0,  0,  0,  0))))
            >>> print(numerix.allclose(cellNormals, mesh._cellNormals, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cellAreaProjections = numerix.array((((0,) * 6, (dy,) * 6, (0,) * 6, (-dy,) * 6),
            ...                                      ((-dx,) * 6, (0,) * 6, (dx,) * 6, (0,) * 6)))

            >>> cellAreaProjections[:, 0] = cellAreaProjections[:, 0] * mesh.cellCenters[0] # doctest: +PROCESSOR_0
            >>> cellAreaProjections[:, 1] = cellAreaProjections[:, 1] * (mesh.cellCenters[0] + mesh.dx / 2.) # doctest: +PROCESSOR_0
            >>> cellAreaProjections[:, 2] = cellAreaProjections[:, 2] * mesh.cellCenters[0] # doctest: +PROCESSOR_0
            >>> cellAreaProjections[:, 3] = cellAreaProjections[:, 3] * (mesh.cellCenters[0] - mesh.dx / 2.) # doctest: +PROCESSOR_0
            >>> print(numerix.allclose(cellAreaProjections, mesh._cellAreaProjections, atol = 1e-10, rtol = 1e-10)) # doctest: +PROCESSOR_0
            True

            >>> cellVertexIDs = MA.masked_array(((5, 6, 7, 9, 10, 11),
            ...                                  (4, 5, 6, 8, 9, 10),
            ...                                  (1, 2, 3, 5, 6, 7),
            ...                                  (0, 1, 2, 4, 5, 6)), -1000)

            >>> print(numerix.allclose(mesh._cellVertexIDs, cellVertexIDs)) # doctest: +PROCESSOR_0
            True

            >>> from fipy.tools import dump
            >>> (f, filename) = dump.write(mesh, extension = '.gz')
            >>> unpickledMesh = dump.read(filename, f)

            >>> print(numerix.allclose(mesh.cellCenters, unpickledMesh.cellCenters))
            True

            >>> faceVertexIDs = [[ 0, 1, 2, 4, 5, 6, 8, 9, 10, 0, 1, 2, 3, 4, 5, 6, 7],
            ...                  [ 1, 2, 3, 5, 6, 7, 9, 10, 11, 4, 5, 6, 7, 8, 9, 10, 11]]
            >>> print(numerix.allequal(mesh.faceVertexIDs, faceVertexIDs)) # doctest: +PROCESSOR_0
            True

            >>> mesh = CylindricalUniformGrid2D(nx=3)
            >>> print(numerix.allequal(mesh._adjacentCellIDs[0],
            ...                        [0, 1, 2, 0, 1, 2, 0, 0, 1, 2])) # doctest: +PROCESSOR_0
            True
            >>> print(numerix.allequal(mesh._adjacentCellIDs[1],
            ...                        [0, 1, 2, 0, 1, 2, 0, 1, 2, 2])) # doctest: +PROCESSOR_0
            True
            >>> faceCellIDs = [[0, 1, 2, 0, 1, 2, 0, 0, 1, 2],
            ...                [-1, -1, -1, -1, -1, -1, -1, 1, 2, -1]]
            >>> print(numerix.allequal(mesh.faceCellIDs.filled(-1),
            ...                        faceCellIDs)) # doctest: +PROCESSOR_0
            True

            >>> mesh = CylindricalUniformGrid2D(ny=3)
            >>> print(numerix.allequal(mesh._adjacentCellIDs[0],
            ...                        [0, 0, 1, 2, 0, 0, 1, 1, 2, 2])) # doctest: +PROCESSOR_0
            True
            >>> print(numerix.allequal(mesh._adjacentCellIDs[1],
            ...                        [0, 1, 2, 2, 0, 0, 1, 1, 2, 2])) # doctest: +PROCESSOR_0
            True
            >>> faceCellIDs = [[0, 0, 1, 2, 0, 0, 1, 1, 2, 2],
            ...                [-1, 1, 2, -1, -1, -1, -1, -1, -1, -1]]
            >>> print(numerix.allequal(mesh.faceCellIDs.filled(-1),
            ...                        faceCellIDs)) # doctest: +PROCESSOR_0
            True

        Following test added to change `nx`, `ny` argument to integer when its a float to prevent
        warnings from the solver.

            >>> from fipy import *
            >>> mesh = CylindricalUniformGrid2D(nx=3, ny=3, dx=1., dy=1.)
            >>> var = CellVariable(mesh=mesh)
            >>> DiffusionTerm().solve(var, solver=DummySolver())

        This test is for https://github.com/usnistgov/fipy/issues/372. Cell
        volumes were being returned as `binOps` rather than arrays.

            >>> m = CylindricalUniformGrid2D(dx=1., dy=1, nx=4, ny=4)
            >>> print(isinstance(m.cellVolumes, numerix.ndarray))
            True
            >>> print(isinstance(m._faceAreas, numerix.ndarray))
            True

        If the above types aren't correct, the divergence operator's value can be a `binOp`

            >>> print(isinstance(CellVariable(mesh=m).arithmeticFaceValue.divergence.value, numerix.ndarray))
            True

        """

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from pysparse.itsolvers import krylov

from fipy.solvers.pysparse.preconditioners import SsorPreconditioner
from fipy.solvers.pysparse.pysparseSolver import PysparseSolver

__all__ = ["LinearPCGSolver"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class LinearPCGSolver(PysparseSolver):
    """

    The `LinearPCGSolver` solves a linear system of equations using the
    preconditioned conjugate gradient method (PCG) with symmetric successive
    over-relaxation (SSOR) preconditioning by default. Alternatively,
    Jacobi preconditioning can be specified through `precon`.
    The PCG method solves systems with
    a symmetric positive definite coefficient matrix.

    The `LinearPCGSolver` is a wrapper class for the the Pysparse_
    `itsolvers.pcg()` and `precon.ssor()` methods.

    .. _Pysparse: http://pysparse.sourceforge.net

    """

    def __init__(self, precon=SsorPreconditioner(), *args, **kwargs):
        """
        Parameters
        ----------
        precon : ~fipy.solvers.pysparse.preconditioners.preconditioner.Preconditioner, optional
        """
        super(LinearPCGSolver, self).__init__(precon=precon, *args, **kwargs)
        self.solveFnc = krylov.pcg

    def _canSolveAsymmetric(self):
        return False



"""Replacement module for NumPy

.. attention::

   This module should be the only place in the code where :mod:`numpy` is
   explicitly imported and you should always import this module and not
   :mod:`numpy` in your own code. The documentation for :mod:`numpy` remains
   canonical for all functions and classes not explicitly documented here.

The functions provided in this module replace and augment the `NumPy` module.
The functions work with `Variables`, arrays or numbers. For example,
create a `Variable`.

   >>> from fipy.variables.variable import Variable
   >>> var = Variable(value=0)

Take the tangent of such a variable. The returned value is itself a
`Variable`.

   >>> v = tan(var)
   >>> v
   tan(Variable(value=array(0)))
   >>> print(float(v))
   0.0

Take the tangent of a int.

   >>> tan(0)
   0.0

Take the tangent of an array.

   >>> print(tan(array((0, 0, 0))))
   [ 0.  0.  0.]

"""
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from builtins import range
from builtins import str
from builtins import zip
__docformat__ = 'restructuredtext'

import numpy as NUMERIX

# On all platforms except Win64, int will be either 32 bit
# or 64 bit depending on the platform and the Python;
# however, on Win64 (and python 64-bit), int will always be
# 32-bit. This is a known issue, and StackOverflow has many
# questions and answers.  However, we still need to do
# something about it, so instead of relying on the default
# mapping, we'll be explicit:

import platform
arch=platform.architecture()[0]
if arch == '32bit':
    INT_DTYPE=NUMERIX.int32
elif arch == '64bit':
    INT_DTYPE=NUMERIX.int64
else:
    raise Exception('Cannot set integer dtype because architecture is unknown.')

from numpy.core import umath
from numpy import newaxis as NewAxis
from numpy import *
try:
    from numpy.core import ma as MA
    numpy_version = 'old'
except ImportError:
    # masked arrays have been moved in numpy 1.1
    from numpy import ma as MA
    numpy_version = 'new'

from fipy.tools import inline

# we want NumPy's __all__, with adjustments
import sys
__all__ = list(sys.modules['numpy'].__dict__.setdefault('__all__', []))
__all__.extend(["NUMERIX", "NewAxis", "MA", "numpy_version"])
__all__.extend(sorted(["getUnit", "put", "reshape", "getShape",
                       "rank", "sum", "isFloat", "isInt", "tostring", "dot",
                       "sqrtDot", "nearest", "allequal", "allclose", "all",
                       "isclose", "take", "indices", "empty", "loadtxt",
                       "savetxt", "L1norm", "L2norm", "LINFnorm", "in1d"],
                      key=str.lower))
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

def _isPhysical(arr):
    """
    Returns `True` if arr is a `Variable` or `PhysicalField`.
    """
    from fipy.variables.variable import Variable
    from fipy.tools.dimensions.physicalField import PhysicalField

    return isinstance(arr, Variable) or isinstance(arr, PhysicalField)

def getUnit(arr):
    if hasattr(arr, "getUnit") and callable(arr.getUnit):
        return arr.unit
    else:
        from fipy.tools.dimensions import physicalField
        return physicalField._unity

def put(arr, ids, values):
    """
    The opposite of `take`.  The values of `arr` at the locations
    specified by `ids` are set to the corresponding value of `values`.

    The following is to test improvements to puts with masked arrays.
    Places in the code were assuming incorrect put behavior.

       >>> maskValue = 999999

       >>> arr = zeros(3, 'l')
       >>> ids = MA.masked_values((2, maskValue), maskValue)
       >>> values = MA.masked_values((4, maskValue), maskValue)
       >>> put(arr, ids, values) ## this should work
       >>> print(arr)
       [0 0 4]

       >>> arr = MA.masked_values((maskValue, 5, 10), maskValue)
       >>> ids = MA.masked_values((2, maskValue), maskValue)
       >>> values = MA.masked_values((4, maskValue), maskValue)
       >>> put(arr, ids, values)
       >>> print(arr) ## works as expected
       [-- 5 4]

       >>> arr = MA.masked_values((maskValue, 5, 10), maskValue)
       >>> ids = MA.masked_values((maskValue, 2), maskValue)
       >>> values = MA.masked_values((4, maskValue), maskValue)
       >>> put(arr, ids, values)
       >>> print(arr) ## should be [-- 5 --] maybe??
       [-- 5 999999]

    """

    if _isPhysical(arr):
        arr.put(ids, values)
    elif MA.isMaskedArray(arr):
        if NUMERIX.sometrue(MA.getmaskarray(ids)):
            if numpy_version == 'old':
                pvalues = MA.array(values, mask=MA.getmaskarray(ids))
            else:
                pvalues = MA.array(values.filled(), mask=MA.getmaskarray(ids))
            MA.put(arr, ids.compressed(), pvalues.compressed())
        else:
            MA.put(arr, ids, values)
    elif MA.isMaskedArray(ids):
        NUMERIX.put(arr, ids.compressed(), MA.array(values, mask=MA.getmaskarray(ids)).compressed())
    else:
        NUMERIX.put(arr, ids, values)

def reshape(arr, shape):
    """
    Change the shape of `arr` to `shape`, as long as the product of all the
    lengths of all the axes is constant (the total number of elements does not
    change).
    """
    if -1 in shape:
        # e.g., NUMERIX.reshape(a, (-1, N)) fails if N == 0
        oldShape = array(getShape(arr))
        oldShape[oldShape == 0] = 1

        if hasattr(shape, 'index'):
            index = shape.index(-1)
        else:
            index = list(shape).index(-1)

        left = shape[:index]
        right = shape[index+1:]
        newShape = array(left + right)
        newShape[newShape == 0] = 1

        shape = left + (oldShape.prod() // newShape.prod(),) + right

    if _isPhysical(arr):
        return arr.reshape(shape)
    elif isinstance(arr, type(array((0)))):
        return NUMERIX.reshape(arr, tuple(shape))
    elif isinstance(arr, type(MA.array((0)))):
        return MA.reshape(arr, shape)
    else:
        return NUMERIX.reshape(array(arr), tuple(shape))
##        raise TypeError, 'cannot reshape object ' + str(arr)

def getShape(arr):
    """
    Return the shape of `arr`

        >>> getShape(1)
        ()
        >>> getShape(1.)
        ()
        >>> from fipy.variables.variable import Variable
        >>> getShape(Variable(1))
        ()
        >>> getShape(Variable(1.))
        ()
        >>> getShape(Variable(1., unit="m"))
        ()
        >>> getShape(Variable("1 m"))
        ()
    """
    if hasattr(arr, "shape"):
        return arr.shape
    elif type(arr) in (type(()), type([])):
        return (len(arr),)
    elif type(arr) in (type(1), type(1.)):
        return ()
    else:
        raise AttributeError("No attribute 'shape'")

def rank(a):
    """
    Get the rank of sequence a (the number of dimensions, not a matrix rank)
    The rank of a scalar is zero.

    .. note::

       The rank of a `MeshVariable` is for any single element. E.g., A
       `CellVariable` containing scalars at each cell, and defined on a 9
       element `Grid1D`, has rank 0. If it is defined on a 3x3 `Grid2D`, it is
       still rank 0.
    """
    if hasattr(a, "rank"):
        return a.rank
    elif hasattr(a, "getRank"):
        return a.getRank()
    else:
        return NUMERIX.rank(a)

def sum(arr, axis=0):
    """
    The sum of all the elements of `arr` along the specified axis.
    """
    if _isPhysical(arr):
        return arr.sum(axis)
    elif isinstance(arr, type(MA.array((0)))):
        return MA.sum(arr, axis)
    else:
        if type(arr) in (float, int) or len(arr) == 0 or 0 in arr.shape:
            return NUMERIX.sum(arr, axis)
        else:
            if axis is None:
                axis = 0
            return NUMERIX.tensordot(NUMERIX.ones(arr.shape[axis], 'l'), arr, (0, axis))

def isFloat(arr):
    if isinstance(arr, NUMERIX.ndarray):
        return NUMERIX.issubclass_(arr.dtype.type, NUMERIX.floating)
    else:
        return NUMERIX.issubclass_(arr.__class__, float)

def isInt(arr):
    if isinstance(arr, NUMERIX.ndarray):
        return NUMERIX.issubclass_(arr.dtype.type, NUMERIX.integer)
    else:
        return NUMERIX.issubclass_(arr.__class__, int)

def tostring(arr, max_line_width=75, precision=8, suppress_small=False, separator=' ', array_output=0):
    r"""
    Returns a textual representation of a number or field of numbers.  Each
    dimension is indicated by a pair of matching square brackets (`[]`), within
    which each subset of the field is output.  The orientation of the dimensions
    is as follows: the last (rightmost) dimension is always horizontal, so that
    the frequent rank-1 fields use a minimum of screen real-estate.  The
    next-to-last dimension is displayed vertically if present and any earlier
    dimension is displayed with additional bracket divisions.

          >>> from fipy import Variable
          >>> print(tostring(Variable((1, 0, 11.2345)), precision=1))
          [  1.    0.   11.2]
          >>> print(tostring(array((1, 2)), precision=5))
          [1 2]
          >>> print(tostring(array((1.12345, 2.79)), precision=2))
          [ 1.12  2.79]
          >>> print(tostring(1))
          1
          >>> print(tostring(array(1)))
          1
          >>> print(tostring(array([1.23345]), precision=2))
          [ 1.23]
          >>> print(tostring(array([1]), precision=2))
          [1]
          >>> print(tostring(1.123456, precision=2))
          1.12
          >>> print(tostring(array(1.123456), precision=3))
          1.123

    Parameters
    ----------
    max_line_width: int
        Maximum number of characters used in a single line.  Default is
        `sys.output_line_width` or 77.
    precision : int
        Number of digits after the decimal point.  Default is
        `sys.float_output_precision` or 8.
    suppress_small : bool
        Whether small values should be suppressed (and output as `0`).
        Default is `sys.float_output_suppress_small` or `False`.
    separator : str
        What character string to place between two numbers.
    array_output : bool
        *unused*
    """

    if _isPhysical(arr):
        return arr.tostring(max_line_width=max_line_width,
                            precision=precision,
                            suppress_small=suppress_small,
                            separator=separator)
    elif isinstance(arr, NUMERIX.ndarray) and arr.shape != ():
        return NUMERIX.array2string(arr,
                                    precision=precision,
                                    max_line_width=max_line_width,
                                    suppress_small=suppress_small,
                                    separator=separator)
    elif isFloat(arr):
        try:
            ## this is for numpy 1.14 and above
            ## why has the interface changed *again*?
            from numpy.core.arrayprint import FloatingFormat
            return FloatingFormat(data=NUMERIX.array((arr,)), precision=precision,
                                  floatmode='maxprec', suppress_small=suppress_small, legacy=False)(arr).strip()
        except ImportError:
            try:
                ## this is for numpy 1.0.4 and above
                ## why has the interface changed again?
                from numpy.core.arrayprint import FloatFormat
                return FloatFormat(NUMERIX.array((arr,)), precision, suppress_small)(arr).strip()
            except ImportError:
                from numpy.core.arrayprint import _floatFormat, _formatFloat
                return _formatFloat(arr, format='%%1.%df' % precision)

    elif isInt(arr):
        try:
            ## this is for numpy 1.7 and above
            ## why has the interface changed again?
            from numpy.core.arrayprint import IntegerFormat
            return IntegerFormat(NUMERIX.array((arr,)))(arr).strip()
        except:
            from numpy.core.arrayprint import _formatInteger
            return _formatInteger(arr, format='%d')
    else:
        raise TypeError('cannot convert ' + str(arr) + ' to string')

#########################
#                       #
#   Vector operations   #
#                       #
#########################

def dot(a1, a2, axis=0):
    """
    return array of vector dot-products of `v1` and `v2`
    for arrays `a1` and `a2` of vectors `v1` and `v2`

    We can't use :func:`numpy.dot` on an array of vectors

    Test that `Variables` are returned as `Variables`.

    >>> from fipy.meshes import Grid2D
    >>> mesh = Grid2D(nx=2, ny=1)
    >>> from fipy.variables.cellVariable import CellVariable
    >>> v1 = CellVariable(mesh=mesh, value=((0, 1), (2, 3)), rank=1)
    >>> v2 = CellVariable(mesh=mesh, value=((0, 1), (2, 3)), rank=1)
    >>> dot(v1, v2)._variableClass
    <class 'fipy.variables.cellVariable.CellVariable'>
    >>> dot(v2, v1)._variableClass
    <class 'fipy.variables.cellVariable.CellVariable'>
    >>> print(rank(dot(v2, v1)))
    0
    >>> print(dot(v1, v2))
    [ 4 10]
    >>> dot(v1, v1)._variableClass
    <class 'fipy.variables.cellVariable.CellVariable'>
    >>> print(dot(v1, v1))
    [ 4 10]
    >>> v3 = array(((0, 1), (2, 3)))
    >>> print(isinstance(dot(v3, v3), type(array(1))))
    1
    >>> print(dot(v3, v3))
    [ 4 10]
    """

    ## have to check MA since MA's have dot() method!!!
    ## and numpy arrays now have a dot() method too!!!!

    def isNumpy(aa):
        return type(aa) in (type(MA.array(0)), type(NUMERIX.array(0)))

    if hasattr(a1, 'dot') and not isNumpy(a1):
        return a1.dot(a2)
    elif hasattr(a2, 'rdot') and not isNumpy(a2):
        return a2.rdot(a1)
    elif hasattr(a2, 'dot') and not isNumpy(a2):
        # dot() is not commutative with tensors, but if there's no
        # rdot(), what else can we do? Just throw an error?
        return a2.dot(a1)
    else:
        return sum(a1*a2, axis)

if inline.doInline:
    def sqrtDot(a1, a2):
        """Return array of square roots of vector dot-products
        for arrays `a1` and `a2` of vectors `v1` and `v2`

        Usually used with `v1==v2` to return magnitude of `v1`.
        """
        from fipy.tools.dimensions import physicalField
        unit1 = unit2 = physicalField._unity

        def dimensionlessUnmasked(a):
            unit = physicalField._unity
            mask = False
            if _isPhysical(a):
                unit = a.inBaseUnits().unit
                a = a.numericValue
            if MA.isMaskedArray(a):
                mask = a.mask
                a = a.filled(fill_value=1)
            if not a.flags['C_CONTIGUOUS']:
                a = a.copy('C')
            return (a, unit, mask)

        a1, unit1, mask1 = dimensionlessUnmasked(a1)
        a2, unit2, mask2 = dimensionlessUnmasked(a2)

        NJ, ni = NUMERIX.shape(a1)
        result1 = NUMERIX.zeros((ni,), 'd')

        inline._runInline("""
            int j;
            result1[i] = 0.;
            for (j = 0; j < NJ; j++)
            {
                result1[i] += a1[i + j * ni] * a2[i + j * ni];
            }
            result1[i] = sqrt(result1[i]);
        """, result1=result1, a1=a1, a2=a2, ni=ni, NJ=NJ)

        if NUMERIX.any(mask1) or NUMERIX.any(mask2):
            result1 = MA.array(result1, mask=NUMERIX.logical_or(mask1, mask2))

        if unit1 != physicalField._unity or unit2 != physicalField._unity:
            from fipy.tools.dimensions.physicalField import PhysicalField
            result1 = PhysicalField(value=result, unit=(unit1 * unit2)**0.5)

        return result1
else:
    def sqrtDot(a1, a2):
        """Return array of square roots of vector dot-products
        for arrays `a1` and `a2` of vectors `v1` and `v2`

        Usually used with `v1==v2` to return magnitude of `v1`.
        """
        ## We can't use Numeric.dot on an array of vectors
        return sqrt(dot(a1, a2))

def nearest(data, points, max_mem=1e8):
    """find the indices of `data` that are closest to `points`

    >>> from fipy import *
    >>> m0 = Grid2D(dx=(.1, 1., 10.), dy=(.1, 1., 10.))
    >>> m1 = Grid2D(nx=2, ny=2, dx=5., dy=5.)
    >>> print(nearest(m0.cellCenters.globalValue, m1.cellCenters.globalValue))
    [4 5 7 8]
    >>> print(nearest(m0.cellCenters.globalValue, m1.cellCenters.globalValue, max_mem=100))
    [4 5 7 8]
    >>> print(nearest(m0.cellCenters.globalValue, m1.cellCenters.globalValue, max_mem=10000))
    [4 5 7 8]
    """
    data = asanyarray(data)
    points = asanyarray(points)

    D = data.shape[0]
    N = data.shape[-1]
    M = points.shape[-1]

    if N == 0:
        return arange(0)

    # given (D, N) data and (D, M) points,
    # break points into (D, C) chunks of points
    # calculate the full factorial (D, N, C) distances between them
    # and then reduce to the indices of the C closest values of data
    # then assemble chunks C into total M closest indices

    # (D, N) -> (D, N, 1)
    data = data[..., newaxis]

    # there appears to be no benefit to taking chunks that use more
    # than about 100 MiB for D x N x C, and there is a substantial penalty
    # for going much above that (presumably due to swapping, even
    # though this is vastly less than the 4 GiB I had available)
    # see ticket:348

    numChunks = int(round(D * N * data.itemsize * M / max_mem + 0.5))

    nearestIndices = empty((M,), dtype=INT_DTYPE)
    for chunk in array_split(arange(points.shape[-1]), numChunks):
        # last chunk can be empty, but numpy (1.5.0.dev8716, anyway)
        # returns array([], dtype=float64), which can't be used for indexing
        chunk = chunk.astype(int)

        # (D, M) -> (D, C)
        chunkOfPoints = points[..., chunk]
        # (D, C) -> (D, 1, C)
        chunkOfPoints = chunkOfPoints[..., newaxis,:]
        # (D, 1, C) -> (D, N, C)
        chunkOfPoints = NUMERIX.repeat(chunkOfPoints, N, axis=1)

#         print "chunkOfPoints size: ", chunkOfPoints.shape, chunkOfPoints.size, chunkOfPoints.itemsize, chunkOfPoints.size * chunkOfPoints.itemsize

        try:
            tmp = data - chunkOfPoints
        except TypeError:
            tmp = data - PhysicalField(chunkOfPoints)

        # (D, N, C) -> (N, C)
        tmp = dot(tmp, tmp, axis=0)

        # (N, C) -> C
        nearestIndices[chunk] = argmin(tmp, axis=0)

    return nearestIndices

def allequal(first, second):
    """
    Returns `true` if every element of `first` is equal to the corresponding
    element of `second`.
    """
    if _isPhysical(first):
        return first.allequal(second)
    elif _isPhysical(second):
        return second.allequal(first)
    else:
        return MA.allequal(first, second)

def allclose(first, second, rtol=1.e-5, atol=1.e-8):
    r"""
    Tests whether or not `first` and `second` are equal, subject to the given
    relative and absolute tolerances, such that::

       |first - second| < atol + rtol * |second|

    This means essentially that both elements are small compared to ``atol`` or
    their difference divided by ``second``'s value is small compared to ``rtol``.
    """
    if _isPhysical(first):
        return first.allclose(other=second, atol=atol, rtol=rtol)
    elif _isPhysical(second):
        return second.allclose(other=first, atol=atol, rtol=rtol)
    else:
        return MA.allclose(first, second, atol=atol, rtol=rtol)

def all(a, axis=None, out=None):
    r"""
    Test whether all array elements along a given axis evaluate to True.

    Parameters
    ----------
    a : array_like
        Input array or object that can be converted to an array.
    axis : int, optional
        Axis along which an logical AND is performed.
        The default (`axis` = `None`) is to perform a logical AND
        over a flattened input array. `axis` may be negative, in which
        case it counts from the last to the first axis.
    out : ndarray, optional
        Alternative output array in which to place the result.
        It must have the same shape as the expected output and
        the type is preserved.

    """
    if _isPhysical(a):
        return a.all(axis=axis)
    else:
        return MA.all(a=a, axis=axis, out=out)

def isclose(first, second, rtol=1.e-5, atol=1.e-8):
    r"""
    Returns which elements of `first` and `second` are equal, subject to the given
    relative and absolute tolerances, such that::

        |first - second| < atol + rtol * |second|

    This means essentially that both elements are small compared to ``atol`` or
    their difference divided by ``second``'s value is small compared to ``rtol``.
    """
    return abs(first - second) < atol + rtol * abs(second)

def take(a, indices, axis=0, fill_value=None):
    """
    Selects the elements of `a` corresponding to `indices`.
    """

    if _isPhysical(a):
        taken = a.take(indices, axis=axis)
    elif isinstance(indices, type(MA.array((0)))):
        ## Replaces `MA.take`. `MA.take` does not always work when
        ## `indices` is a masked array.
        ##
        taken = MA.take(a, MA.filled(indices, 0), axis=axis)

        mask = MA.getmask(indices)

        if mask is not MA.nomask:
            mask = MA.getmaskarray(indices)
            if taken.shape != mask.shape:
                mask = MA.repeat(mask[NewAxis, ...], taken.shape[0], axis=0)
                mask = MA.mask_or(MA.getmask(taken), mask)


        if mask is not MA.nomask:
            taken = MA.array(data=taken, mask=mask)
        else:
            if MA.getmask(taken) is MA.nomask and numpy_version == 'old':
                # numpy 1.1 returns normal array when masked array is filled
                taken = taken.filled()

    elif type(a) in (type(array((0))), type(()), type([])):
        taken = NUMERIX.take(a, indices, axis=axis)
    elif isinstance(a, type(MA.array((0)))):
        taken = MA.take(a, indices, axis=axis)
    else:
        raise TypeError('cannot take from %s object: %s' % (type(a), repr(a)))

    if fill_value is not None and isinstance(taken, type(MA.array((0)))):
        taken = taken.filled(fill_value=fill_value)

    return taken

def L1norm(arr):
    r"""Taxicab or Manhattan norm of `arr`

    :math:`\|\mathtt{arr}\|_1 = \sum_{j=1}^{n} |\mathtt{arr}_j|` is the
    :math:`L^1` norm of :math:`\mathtt{arr}`.

    Parameters
    ----------
    arr : ndarray
    """
    return add.reduce(abs(arr))

def L2norm(arr):
    r"""Euclidean norm of `arr`

    :math:`\|\mathtt{arr}\|_2 = \sqrt{\sum_{j=1}^{n} |\mathtt{arr}_j|^2}` is
    the :math:`L^2` norm of :math:`\mathtt{arr}`.

    Parameters
    ----------
    arr : ndarray
    """
    return sqrt(add.reduce(arr**2))

def LINFnorm(arr):
    """Infinity norm of `arr`

    :math:`\\|\\mathtt{arr}\\|_\\infty = [\\sum_{j=1}^{n}
    |\\mathtt{arr}_j|^\\infty]^\\infty = \\underset{j}{\\max} |\\mathtt{arr}_j|` is the
    :math:`L^\\infty` norm of :math:`\\mathtt{arr}`.

    Parameters
    ----------
    arr : ndarray
    """
    return max(abs(arr))

def _compressIndexSubspaces(index, i, broadcastshape = ()):
    """
    Starting at element `i` in a selection tuple `index`, squeeze out all
    sequential elements that can be broadcast to the same shape.
    """
    skip = 0
    while i + skip < len(index):
        element = index[i + skip]
        if element is newaxis or isinstance(element, slice):
            skip -= 1
            break
        else:
            element = array(element, intp)

            # numpy accepts tuples of lists of floats, but not arrays of
            # floats. This test is more liberal than that, but has the
            # benefit of not being insane. Unfortunately, NumPy will throw
            # errors later, on evaluation, that should ideally be caught here.
            #    raise IndexError "arrays used as indices must be of integer (or boolean) type"

            broadcastshape = _broadcastShape(broadcastshape, element.shape)
            if broadcastshape is None:
                raise ValueError("shape mismatch: objects cannot be broadcast to a single shape")
        skip += 1

    return broadcastshape, skip

def _indexShape(index, arrayShape):
    """
    Given a selection object `index` and an `arrayShape` for the the object to
    be sliced into, return the shape of the result. Return `None` if the
    indexing is not legal.

    "If the length of the selection tuple is larger than N(=`arrayShape.ndim`)
    an error is raised"

        >>> _indexShape(index=(1, 2, 3, 4),
        ...             arrayShape=(10, 20, 30))
        Traceback (most recent call last):
            ...
        IndexError: invalid index

    "All sequences and scalars in the selection tuple are converted to intp
    indexing arrays."

    "All selection tuple objects must be convertible to intp arrays, or slice
    objects, or the Ellipsis (``...``) object"

        >>> _indexShape(index=NUMERIX.index_exp[..., 2, "text"],
        ...             arrayShape=(10, 20, 30, 40, 50))            #doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        ValueError: setting an array element with a sequence.

    .. note::

       The following test should throw::

           Traceback (most recent call last):
               ...
           IndexError: arrays used as indices must be of integer (or boolean) type

       but it's not straightforward to achieve that. Moreover, NumPy is not even
       consistent, accepting a `tuple` or `list` of `float`, but not a `float`
       `array`. If it's absolutely crucial to obtain that result, see the
       comment in `_compressIndexSubspaces()`.

    ..

        >>> ind = zeros((2, 3, 5), float)
        >>> allequal(_indexShape(index=NUMERIX.index_exp[..., ind],
        ...                      arrayShape=(10, 20, 30, 40, 50)),
        ...          (10, 20, 30, 40, 2, 3, 5))
        True

    "Exactly one Ellipsis object will be expanded, any other Ellipsis objects
    will be treated as full slice (`:`) objects. The Ellipsis object is replaced
    with as many full slice (`:`) objects as needed to make the length of the
    selection tuple N."

        >>> _indexShape(index=NUMERIX.index_exp[..., 2, ..., 4],
        ...             arrayShape=(10, 20, 30, 40, 50))
        (10, 20, 40)

    "If the selection tuple is smaller than N, then as many `:` objects as
    needed are added to the end of the selection tuple so that the modified
    selection tuple has length N."

        >>> _indexShape(index=NUMERIX.index_exp[:, 2],
        ...             arrayShape=(10, 20, 30, 40, 50))
        (10, 30, 40, 50)

    "The shape of all the integer indexing arrays must be broadcastable to the
    same shape"

        >>> ind1 = zeros((2, 3, 4), intp)
        >>> ind2 = zeros((2, 3, 5), intp)
        >>> _indexShape(index=NUMERIX.index_exp[:, ind1, ind2],
        ...             arrayShape=(10, 20, 30, 40, 50))
        Traceback (most recent call last):
            ...
        ValueError: shape mismatch: objects cannot be broadcast to a single shape

    "In simple cases (i.e. one indexing array and `N - 1` slice objects) it does
    exactly what you would expect (concatenation of repeated application of
    basic slicing)."

        >>> ind = zeros((2, 3, 4), intp)
        >>> allequal(_indexShape(index=NUMERIX.index_exp[..., ind,:],
        ...                      arrayShape=(10, 20, 30)),
        ...          (10, 2, 3, 4, 30))
        True

    "If the index subspaces are right next to each other, then the broadcasted
    indexing space directly replaces all of the indexed subspaces in X."

        >>> ind1 = zeros((2, 3, 4), intp)
        >>> ind2 = zeros((2, 3, 4), intp)
        >>> allequal(_indexShape(index=NUMERIX.index_exp[:, ind1, ind2],
        ...                      arrayShape=(10, 20, 30, 40, 50)),
        ...          (10, 2, 3, 4, 40, 50))
        True

    "If the indexing subspaces are separated (by slice objects), then the
    broadcasted indexing space is first, followed by the sliced subspace of X."

        >>> allequal(_indexShape(index=NUMERIX.index_exp[:, ind1,:, ind2,:],
        ...                      arrayShape=(10, 20, 30, 40, 50)),
        ...          (2, 3, 4, 10, 30, 50))
        True

    !!! What about Boolean selections ???
    """
    # "when the selection object is not a tuple, it will be referred to as if it
    # had been promoted to a 1-tuple, which will be called the selection tuple"
    if not isinstance(index, tuple):
        if isinstance(index, list):
            index = tuple(index)
        else:
            index = (index,)

    Nnewaxes = len([element for element in index if element is newaxis])
    desiredRank = len(arrayShape) + Nnewaxes

    ellipsislen = 1 + desiredRank - len(index)

    # "Exactly one Ellipsis object will be expanded, any other Ellipsis objects
    # will be treated as full slice (':') objects. The Ellipsis object is replaced
    # with as many full slice (':') objects as needed to make the length of the
    # selection tuple N."
    expanded = ()
    for element in index:
        if element is Ellipsis:
            expanded += (slice(None, None, None),) * ellipsislen
            ellipsislen = 1
        else:
            expanded += (element,)

    if len(expanded) > desiredRank:
        # "If the length of the selection tuple is larger than N (=X.ndim) an error
        # is raised."
        if len(arrayShape) == 0:
            raise IndexError("0-d arrays can't be indexed")
        else:
            raise IndexError("invalid index")
    else:
        # "If the selection tuple is smaller than N, then as many ':' objects as
        # needed are added to the end of the selection tuple so that the modified
        # selection tuple has length N."
        expanded += (slice(None, None, None),) * (desiredRank - len(expanded))

    # "The shape of all the integer indexing arrays must be broadcastable to the
    # same shape"
    broadcasted = ()
    arrayIndices = ()
    arrayindex = None
    broadcastshape = None
    i = 0
    j = 0
    while i < len(expanded):
        element = expanded[i]
        if element is newaxis or isinstance(element, slice):
            broadcasted += (element,)
            if isinstance(element, slice):
                arrayIndices += (j,)
        else:
            if broadcastshape is None:
                arrayindex = i
                broadcastshape, skip = _compressIndexSubspaces(index=expanded, i=i)
            else:
                # we're only in this branch if indexing subspaces are separated
                # by slice objects, so indexing subspace is broadcast first
                arrayindex = 0
                broadcastshape, skip = _compressIndexSubspaces(index=expanded, i=i,
                                                               broadcastshape=broadcastshape)
            i += skip
            j += skip

        i += 1
        if element is not newaxis:
            j += 1

    indexShape = ()
    j = 0
    for element in broadcasted:
        if element is newaxis:
            indexShape += (1,)
        elif isinstance(element, slice):
            start, stop, stride = element.indices(arrayShape[arrayIndices[j]])
            indexShape += ((stop - start) // stride,)
            j += 1
        else:
            raise IndexError("invalid index")

    if arrayindex is not None:
        indexShape = indexShape[:arrayindex] + broadcastshape + indexShape[arrayindex:]

    return indexShape

def _broadcastShapes(shape1, shape2):
    """
    Determine if `shape1` and `shape2` can broadcast to each other, padding if
    necessary, and return their (padded) shapes and the broadcast shape. If the
    shapes cannot broadcast, return a broadcast shape of `None`.

    Broadcasting zero length arrays must also be accounted for.

    >>> _broadcastShapes((1,), (0,))[2]
    (0,)
    >>> _broadcastShapes((2, 0,), (1,))[2]
    (2, 0)

    """

    if len(shape1) > len(shape2):
        shape2 = (1,) * (len(shape1) - len(shape2)) + shape2
    elif len(shape1) < len(shape2):
        shape1 = (1,) * (len(shape2) - len(shape1)) + shape1

    def maxzero(s, o):
        if s == 0 or o == 0:
            return 0
        else:
            return max(s, o)

    if logical_and.reduce([(s == o or s == 1 or o == 1) for s, o in zip(shape1, shape2)]):
        broadcastshape = tuple([maxzero(s, o) for s, o in zip(shape1, shape2)])
    else:
        broadcastshape = None

    return (shape1, shape2, broadcastshape)

def _broadcastShape(shape1, shape2):
    """
    Determine if `shape1` and `shape2` can broadcast to each other, padding if
    necessary, and return the broadcast shape. If the shapes cannot broadcast,
    return `None`.
    """

    shape1, shape2, broadcastshape = _broadcastShapes(shape1, shape2)
    return broadcastshape

if not hasattr(NUMERIX, "in1d"):
    # this handy function was introduced at some point (but it's not in 1.4.1)
    # we define if necessary

    def in1d(ar1, ar2, assume_unique=False):
        """
        Test whether each element of a 1D array is also present in a second array.

        Returns a Boolean array the same length as `ar1` that is True
        where an element of `ar1` is in `ar2` and False otherwise.

        Parameters
        ----------
        ar1 : array_like, shape (M,)
            Input array.
        ar2 : array_like
            The values against which to test each value of `ar1`.
        assume_unique : bool, optional
            If True, the input arrays are both assumed to be unique, which
            can speed up the calculation.  Default is False.

        Returns
        -------
        mask : ndarray of bools, shape(M,)
            The values `ar1[mask]` are in `ar2`.

        See Also
        --------
        numpy.lib.arraysetops : Module with a number of other functions for
                                performing set operations on arrays.

        Notes
        -----
        `in1d` can be considered as an element-wise function version of the
        python keyword `in`, for 1D sequences. ``in1d(a, b)`` is roughly
        equivalent to ``np.array([item in b for item in a])``.

        .. versionadded:: NOT 1.4.0 !!!!

        Examples
        --------
        >>> test = NUMERIX.array([0, 1, 2, 5, 0])
        >>> states = [0, 2]
        >>> mask = in1d(test, states)
        >>> mask
        array([ True, False,  True, False,  True], dtype=bool)
        >>> test[mask]
        array([0, 2, 0])

        """
        if not assume_unique:

            try:
                ar1, rev_idx = NUMERIX.unique(ar1, return_inverse=True)
            except TypeError:
                ar1, rev_idx = NUMERIX.unique1d(ar1, return_inverse=True)

            ar2 = NUMERIX.unique(ar2)

        ar = NUMERIX.concatenate( (ar1, ar2) )
        # We need this to be a stable sort, so always use 'mergesort'
        # here. The values from the first array should always come before
        # the values from the second array.
        order = ar.argsort(kind='mergesort')
        sar = ar[order]
        equal_adj = (sar[1:] == sar[:-1])
        flag = NUMERIX.concatenate( (equal_adj, [False] ) )
        indx = order.argsort(kind='mergesort')[:len( ar1 )]

        if assume_unique:
            return flag[indx]
        else:
            return flag[indx][rev_idx]

def in1dMA(ar1, ar2, assume_unique=False, invert=False):
    """
    Test whether each unmasked element of an array is also present and
    unmasked in a second array.
    The output is always a masked array. See `numpy.in1d` for more details.
    We recommend using :func:`isin` instead of `in1d` for new code.

    Parameters
    ----------
    ar1 : (M,) array_like
        Input array.
    ar2 : array_like
        The values against which to test each value of `ar1`.
    assume_unique : bool, optional
        If True, the input arrays are both assumed to be unique, which
        can speed up the calculation.  Default is False.
    invert : bool, optional
        If True, the values in the returned array are inverted (that is,
        False where an element of `ar1` is in `ar2` and True otherwise).
        Default is False. ``np.in1d(a, b, invert=True)`` is equivalent
        to (but is faster than) ``np.invert(in1d(a, b))``.

    Returns
    -------
    in1d : (M,) ndarray, bool
        The values `ar1[in1d]` are in `ar2`.

    See Also
    --------
    isin       : Version of this function that preserves the shape of ar1.
    numpy.in1d : Equivalent function for ndarrays.

    Notes
    -----
    .. versionadded:: 1.4.0

    Replaces buggy `numpy.ma.in1d`, fixed in
    https://github.com/numpy/numpy/pull/20011.
    """
    ar1, ar2 = MA.asarray(ar1), MA.asarray(ar2)
    m = MA.getmask(ar1)
    res = MA.empty_like(ar1, dtype='bool')
    if ar1.mask is not MA.nomask:
        ar1 = ar1[~ar1.mask]
    if ar2.mask is not MA.nomask:
        ar2 = ar2[~ar2.mask]
    out = NUMERIX.in1d(ar1, ar2, assume_unique=assume_unique,
                      invert=invert)
    if m is not MA.nomask:
        res[~m] = out
    else:
        res[:] = out
    return res

def invert_indices(arr, axis=-1):
    """Invert an index array

    Given an array of indices, return the locations in the array along
    `axis` of each index.

    >>> a = array([[0, 2], [1, 3], [0, 3], [3, 4]])
    >>> print(invert_indices(a, axis=0))
    [[0 2 --]
     [1 -- --]
     [0 -- --]
     [1 2 3]
     [3 -- --]]

    >>> a = MA.masked_values([[0, 1, 0, 3], [2, 3, 3, -1], [-1, 4, -1, -1]], -1)
    >>> print(invert_indices(a, axis=-1))
    [[0 1 0 1 1]
     [2 -- -- 2 --]
     [-- -- -- 3 --]]

    After https://stackoverflow.com/a/59686318/2019542
    """
    from scipy.sparse import coo_matrix
    from scipy.stats.mstats import argstoarray

    fwd = MA.indices(arr.shape)[axis]
    fwd = MA.stack((fwd, arr), axis=-1)
    fwd = MA.reshape(fwd, (-1, 2))
    fwd = MA.compress_rows(fwd)
    rev = coo_matrix(
        (NUMERIX.ones(len(fwd), dtype=int),
         (fwd[..., 0], fwd[..., 1])),
        shape=(arr.shape[axis], max(arr.flat)+1)
    ).tolil().T.rows
    return argstoarray(*rev).swapaxes(0, axis).astype(int)

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()






from __future__ import division
from __future__ import unicode_literals
__all__ = []

from fipy.tools import numerix
from fipy.tools import inline
from fipy.variables.cellVariable import CellVariable

class _AddOverFacesVariable(CellVariable):
    r"""surface integral of `self.faceVariable`, :math:`\phi_f`

    .. math:: \int_S \phi_f\,dS \approx \frac{\sum_f \phi_f A_f}{V_P}

    Returns
    -------
    integral : CellVariable
        volume-weighted sum
    """

    def __init__(self, faceVariable, mesh = None):
        if not mesh:
            mesh = faceVariable.mesh

        CellVariable.__init__(self, mesh, hasOld = 0, elementshape=faceVariable.shape[:-1])
        self.faceVariable = self._requires(faceVariable)

    def _calcValue(self):
        if inline.doInline and self.faceVariable.rank < 2:
            return self._calcValueInline()
        else:
            return self._calcValueNoInline()

    def _calcValueInline(self):

        NCells = self.mesh.numberOfCells
        ids = self.mesh.cellFaceIDs

        val = self._array.copy()

        inline._runInline("""
        int i;

        for(i = 0; i < numberOfCells; i++)
          {
          int j;
          value[i] = 0.;
          for(j = 0; j < numberOfCellFaces; j++)
            {
              // cellFaceIDs can be masked, which caused subtle and
              // unreproducible problems on OS X (who knows why not elsewhere)
              long id = ids[i + j * numberOfCells];
              if (id >= 0) {
                  value[i] += orientations[i + j * numberOfCells] * faceVariable[id];
              }
            }
            value[i] = value[i] / cellVolume[i];
          }
          """,
                          numberOfCellFaces = self.mesh._maxFacesPerCell,
                          numberOfCells = NCells,
                          faceVariable = self.faceVariable.numericValue,
                          ids = numerix.array(ids),
                          value = val,
                          orientations = numerix.array(self.mesh._cellToFaceOrientations),
                          cellVolume = numerix.array(self.mesh.cellVolumes))

        return self._makeValue(value = val)

    def _calcValueNoInline(self):
        ids = self.mesh.cellFaceIDs

        contributions = numerix.take(self.faceVariable, ids, axis=-1)

        # FIXME: numerix.MA.filled casts away dimensions
        s = (numerix.newaxis,) * (len(contributions.shape) - 2) + (slice(0, None, None),) + (slice(0, None, None),)

        faceContributions = contributions * self.mesh._cellToFaceOrientations[s]

        return numerix.tensordot(numerix.ones(faceContributions.shape[-2], 'd'),
                                 numerix.MA.filled(faceContributions, 0.), (0, -2)) / self.mesh.cellVolumes



from __future__ import unicode_literals
import atexit

import pyamgx

from fipy.solvers.pyamgx.pyAMGXSolver import *
from fipy.solvers.pyamgx.linearCGSolver import *
from fipy.solvers.pyamgx.linearGMRESSolver import *
from fipy.solvers.pyamgx.linearFGMRESSolver import *
from fipy.solvers.pyamgx.linearBiCGStabSolver import *
from fipy.solvers.pyamgx.linearLUSolver import *
from fipy.solvers.pyamgx.aggregationAMGSolver import *
from fipy.solvers.pyamgx.classicalAMGSolver import *

pyamgx.initialize()
atexit.register(pyamgx.finalize)

DefaultSolver = LinearCGSolver
DefaultAsymmetricSolver = LinearLUSolver
DummySolver = DefaultSolver
GeneralSolver = LinearLUSolver

__all__ = ["DefaultSolver",
           "DummySolver",
           "DefaultAsymmetricSolver",
           "GeneralSolver"
          ]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

__all__.extend(linearCGSolver.__all__)
__all__.extend(linearFGMRESSolver.__all__)
__all__.extend(linearBiCGStabSolver.__all__)
__all__.extend(linearLUSolver.__all__)
__all__.extend(aggregationAMGSolver.__all__)



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.matrices.scipyMatrix import _ScipyMeshMatrix
from fipy.solvers.solver import Solver
from fipy.tools import numerix

class _ScipySolver(Solver):
    """
    The base `ScipySolver` class.

    .. attention:: This class is abstract. Always create one of its subclasses.
    """

    @property
    def _matrixClass(self):
        return _ScipyMeshMatrix

    def _solve(self):

         if self.var.mesh.communicator.Nproc > 1:
             raise Exception("SciPy solvers cannot be used with multiple processors")

         self.var[:] = numerix.reshape(self._solve_(self.matrix, self.var.ravel(), numerix.array(self.RHSvector)), self.var.shape)
